"""Utility / example functions using new (as of 2.0.0 Docx2Python features)

:author: Shay Hill
:created: 2021-12-21

Docx2Python version two exposes extracted xml in the DocxReader object and has a new
paragraph_styles argument. These functions use these new features as utilities /
examples.
"""

from __future__ import annotations

import re
from pathlib import Path
from typing import Iterator

from lxml.etree import _Element as EtreeElement  # type: ignore

from .iterators import iter_at_depth
from .main import docx2python


def replace_root_text(root: EtreeElement, old: str, new: str) -> None:
    """Replace :old: with :new: in all descendants of :root:

    :param root: an etree element presumably containing descendant text elements
    :param old: text to be replaced
    :param new: replacement text
    """
    for text_elem in (x for x in root.iter() if x.text):
        text_elem.text = (text_elem.text or "").replace(old, new)


def replace_docx_text(
    path_in: Path | str,
    path_out: Path | str,
    *replacements: tuple[str, str],
    html: bool = False,
) -> None:
    """Replace text in a docx file.

    :param path_in: path to input docx
    :param path_out: path to output docx with text replaced
    :param replacements: tuples of strings (a, b) replace a with b for each in docx.
    :param html: respect formatting (as far as docx2python can see formatting)
    """
    reader = docx2python(path_in, html=html).docx_reader
    for file in reader.content_files():
        root = file.root_element
        for replacement in replacements:
            replace_root_text(root, *replacement)
    reader.save(path_out)
    reader.close()


def get_links(path_in: Path | str) -> Iterator[tuple[str, str]]:
    """Iter links inside a docx file as (href, text)

    :param path_in: path to input docx
    :yield: every link in the file as a tuple of (href, text)
    :return: None
    """
    link_pattern = re.compile('<a href="(?P<href>[^"]+)">(?P<text>[^<]+)</a>')
    extraction = docx2python(path_in)
    for run in iter_at_depth(extraction.document_runs, 5):
        match = re.match(link_pattern, run)
        if match:
            href, text = match.groups()
            yield href, text
    extraction.close()


def get_headings(path_in: Path | str) -> Iterator[list[str]]:
    """Iter paragraphs with 'Heading' patagraph_style

    :param path_in: path to input docx
    :yield: every paragraph with 'Heading' paragraph_style as a list of strings
    :return: None

    When docx2python paragraph_styles parameter is set to True, the first run in
    every paragraph will be a paragraph style extracted from the xml, if present.
    Else, paragraphs style will be "".
    """
    heading_pattern = re.compile(r"Heading\d")
    extraction = docx2python(path_in, paragraph_styles=True)
    for par in iter_at_depth(extraction.document_runs, 4):
        if re.match(heading_pattern, par[0]):
            yield par
    extraction.close()



"""Iterate over extracted docx content.

:author: Shay Hill
:created: 6/28/2019

This package extracts docx text as::

    [  # tables
        [  # table
            [  # row
                [  # cell
                    ""  # paragraph
                ]
            ]
        ]
    ]

These functions help manipulate that deep nest without deep indentation.

"""

from __future__ import annotations

from typing import Any, Iterable, Iterator, List, NamedTuple, Sequence, cast

TablesList = List[List[List[List[Any]]]]


class IndexedItem(NamedTuple):
    """The address (indices in a nested list) of an item and the item itself."""

    index: tuple[int, ...]
    value: Any


def enum_at_depth(nested: Sequence[Any], depth: int) -> Iterator[IndexedItem]:
    """Enumerate over a nested sequence at depth.

    :param nested: a (nested) sequence
    :param depth: depth of iteration

        * ``1`` => ``((i,), nested[i])``
        * ``2`` => ``((i, j), nested[:][j])``
        * ``3`` => ``((i, j, k), nested[:][:][k])``
        * ...

    :return: tuples (tuple "address", item)
    :raise ValueError: if depth is less than 1

    >>> sequence = [
    ...     [[["a", "b"], ["c"]], [["d", "e"]]],
    ...     [[["f"], ["g", "h"]]]
    ... ]

    >>> for x in enum_at_depth(sequence, 1): print(x)
    IndexedItem(index=(0,), value=[[['a', 'b'], ['c']], [['d', 'e']]])
    IndexedItem(index=(1,), value=[[['f'], ['g', 'h']]])

    >>> for x in enum_at_depth(sequence, 2): print(x)
    IndexedItem(index=(0, 0), value=[['a', 'b'], ['c']])
    IndexedItem(index=(0, 1), value=[['d', 'e']])
    IndexedItem(index=(1, 0), value=[['f'], ['g', 'h']])

    >>> for x in enum_at_depth(sequence, 3): print(x)
    IndexedItem(index=(0, 0, 0), value=['a', 'b'])
    IndexedItem(index=(0, 0, 1), value=['c'])
    IndexedItem(index=(0, 1, 0), value=['d', 'e'])
    IndexedItem(index=(1, 0, 0), value=['f'])
    IndexedItem(index=(1, 0, 1), value=['g', 'h'])

    >>> for x in enum_at_depth(sequence, 4): print(x)
    IndexedItem(index=(0, 0, 0, 0), value='a')
    IndexedItem(index=(0, 0, 0, 1), value='b')
    IndexedItem(index=(0, 0, 1, 0), value='c')
    IndexedItem(index=(0, 1, 0, 0), value='d')
    IndexedItem(index=(0, 1, 0, 1), value='e')
    IndexedItem(index=(1, 0, 0, 0), value='f')
    IndexedItem(index=(1, 0, 1, 0), value='g')
    IndexedItem(index=(1, 0, 1, 1), value='h')

    >>> list(enum_at_depth(sequence, 5))
    Traceback (most recent call last):
    ...
    TypeError: will not iterate over sequence item

    This error is analogous to the ``TypeError: 'int' object is not iterable`` you
    would see if attempting to enumerate over a non-iterable. In this case,
    you've attempted to enumerate over an item that *may* be iterable, but is not of
    the same type as the ``nested`` sequence argument. This type checking is how we
    can safely descend into a nested list of strings.
    """
    if depth < 1:
        raise ValueError("depth argument must be >= 1")
    argument_type = type(nested)

    def enumerate_next_depth(enumd: Iterable[IndexedItem]) -> Iterator[IndexedItem]:
        """
        Descend into a nested sequence, enumerating along descent

        :param enumd: tuples (tuple of indices, sequences)
        :return: updated index tuples with items from each sequence.
        :raises TypeError: if the sequence is not of the same type as the ``nested``.
            This will happen if you try to iterate into a string in a list of
            strings.
        """
        for index_tuple, sequence in enumd:
            if not isinstance(sequence, argument_type):
                raise TypeError("will not iterate over sequence item")
            for i, item in enumerate(sequence):
                yield IndexedItem(index_tuple + (i,), item)

    depth_n: Iterator[IndexedItem]
    depth_n = (IndexedItem((i,), x) for i, x in enumerate(nested))
    for _ in range(1, depth):
        depth_n = enumerate_next_depth(depth_n)
    return (x for x in depth_n)


def iter_at_depth(nested: Sequence[Any], depth: int) -> Iterator[Any]:
    """
    Iterate over a nested sequence at depth.

    :param nested: a (nested) sequence
    :param depth: depth of iteration

        * ``1`` => ``nested[i]``
        * ``2`` => ``nested[:][j]``
        * ``3`` => ``nested[:][:][k]``
        * ...

    :return: sub-sequences or items in nested

    >>> sequence = [
    ...     [[["a", "b"], ["c"]], [["d", "e"]]],
    ...     [[["f"], ["g", "h"]]]
    ... ]

    >>> for x in iter_at_depth(sequence, 1): print(x)
    [[['a', 'b'], ['c']], [['d', 'e']]]
    [[['f'], ['g', 'h']]]

    >>> for x in iter_at_depth(sequence, 2): print(x)
    [['a', 'b'], ['c']]
    [['d', 'e']]
    [['f'], ['g', 'h']]

    >>> for x in iter_at_depth(sequence, 3): print(x)
    ['a', 'b']
    ['c']
    ['d', 'e']
    ['f']
    ['g', 'h']

    >>> list(iter_at_depth(sequence, 4))
    ['a', 'b', 'c', 'd', 'e', 'f', 'g', 'h']
    """
    return (value for _, value in enum_at_depth(nested, depth))


def iter_tables(tables: TablesList) -> Iterator[list[list[list[Any]]]]:
    """
    Iterate over ``tables[i]``

    Analog of iter_at_depth(tables, 1)

    :param tables: ``[[[["string"]]]]``
    :return: ``tables[0], tables[1], ... tables[i]``
    """
    return iter_at_depth(tables, 1)


def iter_rows(tables: TablesList) -> Iterator[list[list[Any]]]:
    """
    Iterate over ``tables[:][j]``

    Analog of iter_at_depth(tables, 2)

    :param tables: ``[[[["string"]]]]``
    :return: ``tables[0][0], tables[0][1], ... tables[i][j]``
    """
    return iter_at_depth(tables, 2)


def iter_cells(tables: TablesList) -> Iterator[list[Any]]:
    """
    Iterate over ``tables[:][:][k]``

    Analog of iter_at_depth(tables, 3)

    :param tables: ``[[[["string"]]]]``
    :return: ``tables[0][0][0], tables[0][0][1], ... tables[i][j][k]``
    """
    return iter_at_depth(tables, 3)


def iter_paragraphs(tables: TablesList) -> Iterator[str]:
    """
    Iterate over ``tables[:][:][:][l]``

    Analog of iter_at_depth(tables, 4)

    :param tables: ``[[[["string"]]]]``
    :return: ``tables[0][0][0][0], tables[0][0][0][1], ... tables[i][j][k][l]``
    """
    return iter_at_depth(tables, 4)


def enum_tables(tables: TablesList) -> Iterator[IndexedItem]:
    """
    Enumerate over ``tables[i]``

    Analog of enum_at_depth(tables, 1)

    :param tables: ``[[[["string"]]]]``
    :return:
        ``((0, ), tables[0]) ... , ((i, ), tables[i])``
    """
    return enum_at_depth(tables, 1)


def enum_rows(tables: TablesList) -> Iterator[IndexedItem]:
    """
    Enumerate over ``tables[:][j]``

    Analog of enum_at_depth(tables, 2)

    :param tables: ``[[[["string"]]]]``
    :return:
        ``((0, 0), tables[0][0]) ... , ((i, j), tables[i][j])``
    """
    return enum_at_depth(tables, 2)


def enum_cells(tables: TablesList) -> Iterator[IndexedItem]:
    """
    Enumerate over ``tables[:][:][k]``

    Analog of enum_at_depth(tables, 3)

    :param tables: ``[[[["string"]]]]``
    :return:
        ``((0, 0, 0), tables[0][0][0]) ... , ((i, j, k), tables[i][j][k])``
    """
    return enum_at_depth(tables, 3)


def enum_paragraphs(tables: TablesList) -> Iterator[IndexedItem]:
    """
    Enumerate over ``tables[:][:][:][l]``

    Analog of enum_at_depth(tables, 4)

    :param tables: ``[[[["string"]]]]``
    :return:
        ``((0, 0, 0, 0), tables[0][0][0][0]) ... , ((i, j, k, l), tables[i][j][k][l])``
    """
    return enum_at_depth(tables, 4)


def get_text(tables: TablesList) -> str:
    """
    Short cut to pull text from any subset of extracted content.

    :param tables: ``[[[["string"]]]]``
    :return: "string" (all paragraphs in tables joined with '\n\n'
    """
    return "\n\n".join(iter_at_depth(tables, 4))


def get_html_map(tables: TablesList) -> str:
    """
    Create a visual map in html format.

    :param tables: ``[[[["string"]]]]``
    :return: html to show all strings with index tuples

    Create an html string that can be rendered in a browser to show the relative
    location and index tuple of every paragraph in the document.

    * Each table will be a grid of cell boxes, outlined in black. * Each paragraph
    will be prepended with an index tuple. (e.g., ``[[[['text']]]]`` will appear as
    ``(0, 0, 0, 0) text``.
    """

    # prepend index tuple to each paragraph
    tables_4deep = cast(List[List[List[List[str]]]], tables)
    for (i, j, k, l), paragraph in enum_at_depth(tables, 4):
        tables_4deep[i][j][k][l] = " ".join([str((i, j, k, l)), paragraph])

    # wrap each paragraph in <pre> tags
    tables_3deep = cast(List[List[List[str]]], tables_4deep)
    for (i, j, k), cell in enum_at_depth(tables_4deep, 3):
        cell = (str(x) for x in cell)
        tables_3deep[i][j][k] = "".join([f"<pre>{x}</pre>" for x in cell])

    # wrap each cell in <td> tags
    tables_2deep = cast(List[List[str]], tables_3deep)
    for (i, j), row in enum_at_depth(tables_3deep, 2):
        tables_2deep[i][j] = "".join([f"<td>{x}</td>" for x in row])

    # wrap each row in <tr> tags
    tables_1deep = cast(List[str], tables_2deep)
    for (i,), table in enum_at_depth(tables_2deep, 1):
        tables_1deep[i] = "".join(f"<tr>{x}</tr>" for x in table)

    # wrap each table in <table> tags
    tables_ = "".join([f'<table border="1">{x}</table>' for x in tables_1deep])

    return "<html><body>" + tables_ + "</body></html>"



""" Merge runs with identical formatting.

:author: Shay Hill
:created: 12/13/2021

Join consecutive xml runs with identical formatting. See docstring for ``merge_elems``.
"""
from __future__ import annotations

import functools
from itertools import groupby
from typing import TYPE_CHECKING

from lxml.etree import _Element as EtreeElement  # type: ignore

from .attribute_register import RELS_ID, Tags, has_content
from .text_runs import get_html_formatting

if TYPE_CHECKING:
    from .docx_reader import File

# identify tags that will be merged together (if formatting is equivalent)
_MERGEABLE_TAGS = {Tags.RUN, Tags.HYPERLINK, Tags.TEXT, Tags.TEXT_MATH}


def _elem_key(file: File, elem: EtreeElement) -> tuple[str, str, list[str]]:
    """
    Enough information to tell if two elements are more-or-less identically formatted.

    :param elem: any element in an xml file.
    :return: A summary of attributes (if two adjacent elements return the same key,
        they are considered mergeable). Only used to merge elements, so returns None
        if element tags are not in _MERGEABLE_TAGS.

    Ignore text formatting differences if consecutive link elements point to the same
    address. Always join these.

    Docx2Text joins consecutive runs and links of the same style. Comparing two
    elem_key return values will tell you if
        * elements are the same type
        * link rels ids reference the same link
        * run styles are the same (as far as docx2python understands them)

    Elem rId attributes are replaced with rId['Target'] because different rIds can
    point to identical targets. This is important for hyperlinks, which can look
    different but point to the same address.

    """
    tag = elem.tag
    if tag not in _MERGEABLE_TAGS:
        return tag, "", []

    # always join links pointing to the same address
    rels_id = elem.attrib.get(RELS_ID)
    if rels_id:
        return tag, str(file.rels[str(rels_id)]), []

    return tag, "", get_html_formatting(elem, file.context.xml2html_format)


def merge_elems(file: File, tree: EtreeElement) -> None:
    """
    Recursively merge duplicate (as far as docx2python is concerned) elements.

    :param file: File instancce
    :param tree: root_element from an xml in File instance
    :effects: Merges consecutive elements if tag, attrib, and style are the same

    There are a few ways consecutive elements can be "identical":
        * same link
        * same style

    Often, consecutive, "identical" elements are written as separate elements,
    because they aren't identical to Word. Word keeps track of revision history,
    spelling errors, etc., which are meaningless to docx2python.

    <w:p>
        <w:hyperlink r:id="rId7">  <!-- points to http://www.shayallenhill.com -->
            <w:r>
                <w:t>hy</w:t>
            </w:r>
        </w:hyperlink>
        <w:proofErr/>  <!-- docx2python will ignore this proofErr -->
        <w:hyperlink r:id="rId8">  <!-- points to http://www.shayallenhill.com -->
            <w:r>
                <w:t>per</w:t>
            </w:r>
        </w:hyperlink>
        <w:hyperlink r:id="rId9">  <!-- points to http://www.shayallenhill.com -->
            <w:r w:rsid="asdfas">  <!-- docx2python will ignore this rsid -->
                <w:t>link</w:t>
            </w:r>
        </w:hyperlink>
    </w:p>

    Docx2python condenses the above to (by merging links)

    <w:p>
        <w:hyperlink r:id="rId7">  <!-- points to http://www.shayallenhill.com -->
            <w:r>
                <w:t>hy</w:t>
            </w:r>
            <w:r>
                <w:t>per</w:t>
            </w:r>
            <w:r w:rsid="asdfas">  <!-- docx2python will ignore this rsid -->
                <w:t>link</w:t>
            </w:r>
        </w:hyperlink>
    </w:p>

    Then to (by merging runs)

    <w:p>
        <w:hyperlink r:id="rId7">  <!-- points to http://www.shayallenhill.com -->
            <w:r>
                <w:t>hy</w:t>
                <w:t>per</w:t>
                <w:t>link</w:t>
            </w:r>
        </w:hyperlink>
    </w:p>

    Then finally to (by merging text)

    <w:p>
        <w:hyperlink r:id="rId7">  <!-- points to http://www.shayallenhill.com -->
            <w:r>
                <w:t>hyperlink</w:t>
            </w:r>
        </w:hyperlink>
    </w:p>

    This function only merges runs, text, and hyperlinks, because merging paragraphs
    or larger elements would ignore information docx2python DOES want to preserve.

    Filter out non-content items so runs can be joined even
    """

    file_elem_key = functools.partial(_elem_key, file)

    elems = [x for x in tree if has_content(x)]
    runs = [list(y) for _, y in groupby(elems, key=file_elem_key)]

    for run in (x for x in runs if len(x) > 1 and x[0].tag in _MERGEABLE_TAGS):
        if run[0].tag in {Tags.TEXT, Tags.TEXT_MATH}:
            run[0].text = "".join(x.text or "" for x in run)
        for elem in run[1:]:
            for e in elem:
                run[0].append(e)
            tree.remove(elem)

    for branch in tree:
        merge_elems(file, branch)



""" Form checkboxes, dropdowns, and other non-text elements visible in Word.

:author: Shay Hill
:created: 6/17/2020

Word represents some special characters as non-text elements (e.g., checkBox). These
functions examine these elements to infer suitable text replacements.

This file references "\u2610" and "\u2612" a few times. These are open and
crossed-out checkboxes. Pypi doesn't like them in my file, so I have to reference
them by their escape sequences.
"""

from contextlib import suppress
from typing import Union

from lxml.etree import _Element as EtreeElement  # type: ignore

from .namespace import qn


def get_checkBox_entry(checkBox: EtreeElement) -> str:
    """Create text representation for a checkBox element.

    :param checkBox: a checkBox xml element
    :return:
        1. attempt to get ``checked.w:val`` and return "\u2610" or "\u2612"
        2. attempt to get ``default.w:val`` and return "\u2610" or "\u2612"
        3. return ``--checkbox failed--``

    Docx xml has at least two types of checkbox elements::

        1. ``checkBox`` can only be checked when the form is locked. These do not
        contain a text element, so this function is needed to select one from the
        ``w:checked`` or ``w:default`` sub-elements.

        2. ``checkbox`` can be checked any time. Prints text as "\u2610" or "\u2612".
        Docx2Python can safely ignore this second type, as there will be a <w:t>
        element inside with a checkbox character.

    <w:checkBox>
        <w:sizeAuto/>
        <w:default w:val="1"/>
        <w:checked w:val="0"/>
    </w:checkBox>

    If the ``checked`` attribute is absent, return the default
    If the ``checked`` attribute is present, but not w:val is given, return unchecked
    """

    def get_wval() -> Union[str, None]:
        """Get the value of the ``w:val`` attribute of the ``checked`` element.

        :return: the value of the ``w:val`` attribute of the ``checked`` element
        """
        with suppress(StopIteration):
            checked = next(checkBox.iterfind(qn("w:checked")))
            return str(checked.attrib.get(qn("w:val")) or "1")
        with suppress(StopIteration, KeyError):
            default = next(checkBox.iterfind(qn("w:default")))
            return str(default.attrib[qn("w:val")])
        return None

    return {"0": "\u2610", "1": "\u2612", None: "----checkbox failed----"}[get_wval()]


def get_ddList_entry(ddList: EtreeElement) -> str:
    """Get only the selected string of a dropdown list.

    :param ddList: a dropdown-list element
    :return: w:listEntry value of input element.

    <w:ddList>
        <w:result w:val="1"/>
        <w:listEntry w:val="selection 1"/>
        <w:listEntry w:val="selection 2"/>
    </w:ddList>

    <w:result w:val="0"/> might be missing when selection is "0"
    """
    list_entries = [
        x.attrib.get(qn("w:val")) for x in ddList.findall(qn("w:listEntry"))
    ]
    try:
        result = next(ddList.iterfind(qn("w:result")))
        list_index = int(result.attrib[qn("w:val")])
    except (StopIteration, KeyError):
        list_index = 0
    return str(list_entries[list_index])



""" The tags and attributes docx2python knows how to handle.

:author: Shay Hill
:created: 3/18/2021

A lot of the information in a docx file isn't text or text attributes. Docx files
record spelling errors, revision history, etc. Docx2Python will ignore (by design)
much of this.
"""
from enum import Enum
from typing import Callable, Iterator, NamedTuple, Optional

from lxml.etree import _Element as EtreeElement  # type: ignore

from .namespace import qn


class HtmlFormatter(NamedTuple):
    """
    The information needed to group and format html tags.

    If a text run has multiple span attributes, Docx2Python does not open multiple
    span elements.

    ``<span style="font-size:12"><span style="font-weight:bold"> ...``

    Instead, these elements are first grouped by self.container, then by
    self.property, then printed all together.

    ``<span style="font-size:12;font-weight:bold">``

    This makes a more-readable text extraction, but it does involve passing these
    HtmlFormatter instances around A LOT.

    Formatting in the xml file will appear as child elements of an rPr or pPr element:

    ``<w:sz w:val="32"/>``

    self.formatter will be a function that takes the element tag name (here ``sz``)
    and element ``w:val`` attribute (here ``"32"``).
    """

    formatter: Callable[[str, str], str] = lambda tag, val: tag
    container: Optional[str] = None  # e.g., 'span'
    property_: Optional[str] = None  # e.g., 'style'


# An HtmlFormatter instance for every xml format Docx2Python recognizes.
# This mapping can be extended from outside by
#
#     import docx2python
#     docx2python.attribute_register.xml2html_formatter[xml_format] =
#         docx2python.attribute_register.HtmlFormatter(args)
XML2HTML_FORMATTER = {
    "b": HtmlFormatter(),
    "i": HtmlFormatter(),
    "u": HtmlFormatter(),
    "strike": HtmlFormatter(lambda tag, val: "s"),
    "vertAlign": HtmlFormatter(lambda tag, val: val[:3]),  # subscript and superscript
    "smallCaps": HtmlFormatter(
        lambda tag, val: "font-variant:small-caps", "span", "style"
    ),
    "caps": HtmlFormatter(lambda tag, val: "text-transform:uppercase", "span", "style"),
    "highlight": HtmlFormatter(
        lambda tag, val: f"background-color:{val}", "span", "style"
    ),
    "sz": HtmlFormatter(lambda tag, val: f"font-size:{val}pt", "span", "style"),
    "color": HtmlFormatter(lambda tag, val: f"color:{val}", "span", "style"),
    "Heading1": HtmlFormatter(lambda tag, val: "h1"),
    "Heading2": HtmlFormatter(lambda tag, val: "h2"),
    "Heading3": HtmlFormatter(lambda tag, val: "h3"),
    "Heading4": HtmlFormatter(lambda tag, val: "h4"),
    "Heading5": HtmlFormatter(lambda tag, val: "h5"),
    "Heading6": HtmlFormatter(lambda tag, val: "h6"),
}


class Tags(str, Enum):
    """
    These are the tags that provoke some action in docx2python.
    """

    BODY = qn("w:body")
    BR = qn("w:br")
    DOCUMENT = qn("w:document")
    ENDNOTE = qn("w:endnote")
    ENDNOTE_REFERENCE = qn("w:endnoteReference")
    FOOTNOTE = qn("w:footnote")
    FOOTNOTE_REFERENCE = qn("w:footnoteReference")
    FORM_CHECKBOX = qn("w:checkBox")
    FORM_DDLIST = qn("w:ddList")  # drop-down form
    HYPERLINK = qn("w:hyperlink")
    IMAGE = qn("a:blip")
    IMAGEDATA = qn("v:imagedata")
    MATH = qn("m:oMath")
    PARAGRAPH = qn("w:p")
    PAR_PROPERTIES = qn("w:pPr")
    RUN = qn("w:r")
    RUN_PROPERTIES = qn("w:rPr")
    SYM = qn("w:sym")
    TAB = qn("w:tab")
    TABLE = qn("w:tbl")
    TABLE_CELL = qn("w:tc")
    TABLE_ROW = qn("w:tr")
    TEXT = qn("w:t")
    TEXT_MATH = qn("m:t")


# elem.attrib key for relationship ids. These can find the information they reference by
# ``file_instance.rels[elem.attrib[RELS_ID]]``
RELS_ID = qn("r:id")

_CONTENT_TAGS = set(Tags) - {Tags.RUN_PROPERTIES, Tags.PAR_PROPERTIES}


def has_content(tree: EtreeElement) -> Optional[str]:
    """
    Does the element have any descendent content elements?

    :param tree: xml element
    :return: first content tag found or None if no content tags are found

    This is to check for text in any skipped elements.

    Docx2Python ignores spell check, revision, and other elements. This function checks
    that no content (paragraphs, run, text, link, ...) is contained in children of any
    ignored elements.

    If no content is found, the element can be safely ignored going forward.
    """

    def iter_content(tree_: EtreeElement) -> Iterator[str]:
        """Yield all content elements in tree

        :param tree_: xml element
        :yield: child content elements
        :return: None
        """
        if tree_.tag in _CONTENT_TAGS:
            yield tree_.tag
        for branch in tree_:
            yield from iter_content(branch)

    return next(iter_content(tree), None)



"""Numbering formats for converted XML lists.

:author: Shay Hill
:created: 6/26/2019

I don't want to add non-ascii text to a potentially ascii-only file, so all bullets
are '--' and Roman numerals stop at 3999.

Doesn't capture formatting like 1.1.1 or b) or (ii). Only the six basic formats are
covered::

    -- bullet
    1  decimal
    a  lowerLetter
    A  upperLetter
    i  lowerRoman
    I  upperRoman
"""

from string import ascii_lowercase

# Subs to convert any number of i's to a proper Roman numeral
# fmt=off
ROMAN_SUBS = [
    ("iiiii", "v"),  # 1+1+1+1+1 -> 5
    ("vv", "x"),  # 5+5 -> 10
    ("xxxxx", "l"),  # 10+10+10+10 -> 50
    ("ll", "c"),  # 50+50 -> 100
    ("ccccc", "d"),  # 100+100+100+100+100 -> 500
    ("dd", "m"),  # 500+500 -> 1000
    ("iiii", "iv"),  # 1+1+1+1 -> 4
    ("viv", "ix"),  # 5+4 -> 9
    ("xxxx", "xl"),  # 10+10+10+10 -> 40
    ("lxl", "xc"),  # 50+40 -> 90
    ("cccc", "cd"),  # 100+100+100+100 -> 40
    ("dcd", "cm"),  # 500+400 -> 900
]
# fmt=on


def lower_letter(n: int) -> str:
    """
    Convert a positive integer to a string of letters representing base 26.

    :param n: any positive integer
    :return: the kind of "numbering" used for numbered lists and excel columns.
        (a, b, c ... aa, ab ...) Zero is undefined.
    :raise ValueError: if n is not a positive integer

        >>> lower_letter(1)
        'a'
        >>> lower_letter(26)
        'z'
        >>> lower_letter(27)
        'aa'
    """
    if n < 1:
        raise ValueError("0 and <1 are not defined for this numbering")
    result = ""
    while n:
        n, remainder = divmod(n - 1, 26)
        result = ascii_lowercase[remainder] + result
    return result


def upper_letter(n: int) -> str:
    """Get int as an upprecase letter.

    :param n: any positive integer
    :return: the kind of "numbering" used for numbered lists and excel columns.
    """
    return lower_letter(n).upper()


def lower_roman(n: int) -> str:
    """
    Convert a positive integer to a lowercase Roman numeral

    :param n: any positive integer
    :return: Roman number equivalent of n
    :raise ValueError: if n is not a positive integer

        >>> lower_roman(1)
        'i'
        >>> lower_roman(9)
        'ix'
        >>> lower_roman(44)
        'xliv'

    Numbers greater than 3999 can be expressed with a bar over the number. The bar
    means "times 1000" (e.g., iv with a bar over it would be 4000).

    It'll never happen in this project, and I don't want to add non-ascii to what
    might be a pure ascii file, so this function will keep adding 'm' to as many
    thousand as you'd like.

        >>> lower_roman(10000)
        'mmmmmmmmmm'
    """
    if n < 1:
        raise ValueError(f"the Romans hadn't figured out {n}")
    result = "i" * n
    for pattern, replacement in ROMAN_SUBS:
        result = result.replace(pattern, replacement)
    return result


def upper_roman(n: int) -> str:
    """Get int as an uppercase Roman numeral.

    :param n: any positive integer
    :return: Roman number equivalent of n
    """
    return lower_roman(n).upper()


def decimal(n: int) -> str:
    """Get int as a decimal number string

    :param n: any integer
    :return: string such that int(decimal(n)) == n
    """
    return str(n)


def bullet(_: int = 0) -> str:
    """Get the string we're using to replace bullets.

    :return: the string we're using to replace bullets.
    """
    return "--"



""" Hold and decode docx internal xml files.

:author: Shay Hill
:created: 3/18/2021

See the docx file structure in ``README_DOCX_FILE_STRUCTURE.md``. Each file in that
structure can be stored as a ``File`` instance, though not all will be stored through
the typical docx2python progression.

The ``File`` class is designed to hold and decode xml files with content (text).
Several of the xml files in a docx do not contain content. These contain numbering
formats, font information, rId-lookup tables, and other. ``File`` instances will hold
these as well, though they will not have ``rels`` or ``content`` attributes. Will
return an empty dictionary or empty list if asked.

Some of these non-content files are shared between between . The substance of these
files is accessible through the ``DocxContent`` class. This class holds file
instances and decodes shared non-content in a docx file structure.
"""

from __future__ import annotations

import copy
import os
import pathlib
import zipfile
from contextlib import suppress
from dataclasses import dataclass
from io import BytesIO
from operator import attrgetter
from pathlib import Path
from typing import Any
from warnings import warn

from lxml import etree
from lxml.etree import _Element as EtreeElement  # type: ignore

from .attribute_register import XML2HTML_FORMATTER
from .docx_context import collect_numFmts, collect_rels
from .docx_text import get_text
from .merge_runs import merge_elems

CONTENT_FILE_TYPES = {"officeDocument", "header", "footer", "footnotes", "endnotes"}


@dataclass
class File:
    """The attribute dict of a file in the docx, plus cached data

    The docx lists internal files in various _rels files. Each will be specified with a
    dict of, e.g.::

        {
            'Id': 'rId8',
            'Type': 'http://schemas.openxmlformats.org/.../relationships/header',
            'Target': 'header1.xml'
        }

    This isn't quite enough to infer the structure of the docx. You'll also need to
    know the directory where each attribute dict was found::

        'dir': 'word/_rels'

    That's the starting point for these instances. Other attributes are inferred or
    created at runtime.
    """

    def __init__(self, context: DocxReader, attribute_dict: dict[str, str]) -> None:
        """
        Point to container DocxContext instance and store attributes as properties.

        :param context: The DocxContent object holding this instance
        :param attribute_dict: Attributes of this file found in the rels, plus 'dir' as
            described above.
        """
        self.context = context
        self.Id = str(attribute_dict["Id"])
        self.Type = os.path.basename(attribute_dict["Type"])
        self.Target = attribute_dict["Target"]
        self.dir = attribute_dict["dir"]

        # cached_properties
        self.__path: None | str = None
        self.__rels_path: None | str = None
        self.__rels: None | dict[str, str] = None
        self.__root_element: None | EtreeElement = None

    def __repr__(self) -> str:
        """File with self.path

        :return: String representation
        """
        return f"File({self.path})"

    @property
    def path(self) -> str:
        """Infer path/to/xml/file from instance attributes

        :return: path to xml file

        This will take the information in a file specification (from one of the rels
        files, e.g., {Id:'  ', Type:'  ' Target:'  ', dir:'  '}) and infer a path to
        the specified xml file.

        E.g.,
        from     self.dir = '_rels'       self.Target = 'word/document.xml
                    dirname ''          +       dirname 'word/'
                                        +       filename =   'document.xml'
        return `word/document`

        E.g.,
        from     self.dir = 'word/_rels'       self.Target = 'header1.xml
                    dirname 'word'      +            dirname ''
                                        +       filename =   'header1.xml'
        return `word/header1.xml`
        """
        if self.__path is not None:
            return self.__path

        dirs = [os.path.dirname(x) for x in (self.dir, self.Target)]
        dirname = "/".join([x for x in dirs if x])
        filename = os.path.basename(self.Target)
        self.__path = "/".join([dirname, filename])
        return self.__path

    @property
    def _rels_path(self) -> str:
        """Infer path/to/rels from instance attributes.

        :return: path to rels (which may not exist)

        Every content file (``document.xml``, ``header1.xml``, ...) will have its own
        ``.rels`` file--if any relationships are defined.

        The path inferred here may not exist.

        E.g.,
        from     self.dir = '_rels'       self.Target = 'word/document.xml
                    dirname ''          +       dirname 'word/'
                                        +       filename =   'document.xml'
        return `word/_rels/document.xml.rels`

        E.g.,
        from     self.dir = 'word/_rels'       self.Target = 'header1.xml
                    dirname 'word'      +            dirname ''
                                        +       filename =   'header1.xml'
        return `word/_rels/header1.xml.rels`
        """
        if self.__rels_path is not None:
            return self.__rels_path
        dirname, filename = os.path.split(self.path)
        self.__rels_path = "/".join([dirname, "_rels", filename + ".rels"])
        return self.__rels_path

    @property
    def rels(self) -> dict[str, str]:
        """rIds mapped to values

        :return: dict of rIds mapped to values

        Each content file.xml will have a file.xml.rels file--if relationships are
        defined. Inside file.xml, values defined in the file.xml.rels file may be
        referenced by their rId numbers.

        :return: Contents of the file.xml.rels file with reference rId numbers. These
        refer to values defined in the file.xml.rels file:

        E.g.::

        {
            "rId3": "webSettings.xml",
            "rId2": "settings.xml",
            "rId1": "styles.xml",
            "rId6": "theme/theme1.xml",
            "rId5": "fontTable.xml",
            "rId4": "https://www.shayallenhill.com/",
        }

        Not every xml file with have a rels file. Return an empty dictionary if the
        rels file is not found.
        """
        if self.__rels is not None:
            return self.__rels

        try:
            unzipped = self.context.zipf.read(self._rels_path)
            tree = etree.fromstring(unzipped)
            self.__rels = {str(x.attrib["Id"]): str(x.attrib["Target"]) for x in tree}
        except KeyError:
            self.__rels = {}
        return self.__rels

    @property
    def root_element(self) -> EtreeElement:
        """Root element of the file.

        :return: Root element of the file.

        Try to merge consecutive, duplicate (except text) elements in content files.
        See documentation for ``merge_elems``. Warn if ``merge_elems`` fails.
        (I don't think it will fail).
        """
        if self.__root_element is not None:
            return self.__root_element

        root = etree.fromstring(self.context.zipf.read(self.path))
        if self.Type in CONTENT_FILE_TYPES:
            root_ = copy.copy(root)
            try:
                merge_elems(self, root)
            except Exception as ex:
                warn(
                    "Attempt to merge consecutive elements in "
                    + f"{self.context.docx_filename} {self.path} resulted in "
                    + f"{repr(ex)}. Moving on."
                )
                self.__root_element = root_
        self.__root_element = root
        return self.__root_element

    @property
    def content(self) -> list[list[list[list[str]]]]:
        """Text extracted into a 5-layer-deep nested list of strings.

        :return: Text extracted into a 5-layer-deep nested list of strings.
        """
        return get_text(self)

    def get_content(
        self, root: EtreeElement | None = None
    ) -> list[list[list[list[str]]]]:
        """
        The same content as property 'content' with optional given root.

        :param root: Extract content of file from root down.
            If root is not given, return full content of file.
        :return: Text extracted into a 5-layer-deep nested list of strings.
        """
        return get_text(self, root)


@dataclass
class DocxReader:
    """
    Hold File instances and decode information shared between them (e.g., numFmts)
    """

    def __init__(
        self,
        docx_filename: Path | str | BytesIO,
        html: bool = False,
        paragraph_styles: bool = False,
        duplicate_merged_cells: bool = False,
    ):
        self.docx_filename = docx_filename
        self.do_pStyle = paragraph_styles
        self.duplicate_merged_cells = duplicate_merged_cells

        if html:
            self.xml2html_format = XML2HTML_FORMATTER
        else:
            self.xml2html_format = {}

        # cached properties and a flag (__closed)
        self.__zipf: None | zipfile.ZipFile = None
        self.__files: None | list[File] = None
        self.__numId2NumFmts: None | dict[str, list[str]] = None
        self.__closed = False

    @property
    def zipf(self) -> zipfile.ZipFile:
        """
        Entire docx unzipped into bytes.

        :return: Entire docx unzipped into bytes.
        :raise ValueError: If DocxReader instance has been closed
        """
        if self.__closed:
            raise ValueError("DocxReader instance has been closed.")
        if self.__zipf is None:
            self.__zipf = zipfile.ZipFile(self.docx_filename)
            return self.__zipf
        assert self.__zipf is not None
        return self.__zipf

    def close(self):
        """Close the zipfile, set __closed flag to True."""
        if self.__zipf is not None and self.__zipf.fp:
            self.__zipf.close()
        self.__closed = True

    def __enter__(self) -> DocxReader:
        """Do nothing. The zipfile will open itself when needed.

        :return: self
        """
        return self

    def __exit__(
        self,
        exc_type: Any,  # None | Type[Exception], but py <= 3.9 doesn't like it.
        exc_value: Any,  # None | Exception, but py <= 3.9 doesn't like it.
        exc_traceback: Any,  # None | TracebackType, but py <= 3.9 doesn't like it.
    ):
        """Close the zipfile.

        :param exc_type: Python internal use
        :param exc_value: Python internal use
        :param exc_traceback: Python internal use
        """
        self.close()

    @property
    def files(self) -> list[File]:
        """
        Instantiate a File instance for every content file.

        :return: List of File instances, one per content file.
        """
        if self.__files is not None:
            return self.__files

        files: list[File] = []
        for k, v in collect_rels(self.zipf).items():
            files += [File(self, {**x, "dir": os.path.dirname(k)}) for x in v]
        self.__files = files
        return self.__files

    @property
    def numId2numFmts(self) -> dict[str, list[str]]:
        """
        numId referenced in xml to list of numFmt per indentation level

        :return: numId referenced in xml to list of numFmt per indentation level

        See docstring for collect_numFmts

        Returns an empty dictionary if word/numbering.xml cannot be found.
        Ultimately, this will result in any lists (there should NOT be any lists if
        there is no word/numbering.xml) being "numbered" with "--".
        """
        if self.__numId2NumFmts is not None:
            return self.__numId2NumFmts

        try:
            numFmts_root = etree.fromstring(self.zipf.read("word/numbering.xml"))
            self.__numId2NumFmts = collect_numFmts(numFmts_root)
        except KeyError:
            self.__numId2NumFmts = {}
        return self.__numId2NumFmts

    def file_of_type(self, type_: str) -> File:
        """
        Return file instance attrib Type='http://.../type_'. Warn if more than one.

        :param type_: this package looks for any of
            ("header", "officeDocument", "footer", "footnotes", "endnotes")
            You can try others.
        :return: File instance of the requested type
        :raise KeyError: if no file of the requested type is found
        """
        files_of_type = self.files_of_type(type_)
        if len(files_of_type) > 1:
            warn("Multiple files of type '{type_}' found. Returning first.")
        try:
            return files_of_type[0]
        except IndexError as exc:
            raise KeyError(
                f"There is no item of type '{type_}' "
                + "in the {self.docx_filename} archive"
            ) from exc

    def files_of_type(self, type_: str | None = None) -> list[File]:
        """
        File instances with attrib Type='http://.../type_'

        :param type_: this package looks for any of
            ("header", "officeDocument", "footer", "footnotes", "endnotes")
            You can try others. If argument is None (default), returns all content file
            types.
        :return: File instances of the requested type, sorted by path
        """
        if type_ is None:
            types = CONTENT_FILE_TYPES
        else:
            types = {type_}

        return sorted(
            (x for x in self.files if x.Type in types), key=attrgetter("path")
        )

    def content_files(self) -> list[File]:
        """
        Content files (contain displayed text) inside the docx.

        :return: File instances of context files, sorted by path
        """
        return self.files_of_type()

    def save(self, filename: Path | str) -> None:
        """
        Save the (presumably altered) xml.

        :param filename: path to output file (presumably *.docx)

        xml (root_element) attributes are cached, so these can be altered and saved.
        This allows saving a copy of the input docx after the ``merge_elems`` operation.
        Also allows some light editing like search and replace.
        """
        content_files = [x for x in self.files if x.Type in CONTENT_FILE_TYPES]
        with zipfile.ZipFile(f"{filename}", mode="w") as zout:
            _copy_but(self.zipf, zout, {x.path for x in content_files})
            for file in content_files:
                zout.writestr(file.path, etree.tostring(file.root_element))

    def pull_image_files(self, image_directory: str | None = None) -> dict[str, bytes]:
        """
        Copy images from zip file.

        :param image_directory: optional destination for copied images
        :return: Image names mapped to images in binary format.

            To write these to disc::

                with open(key, 'wb') as file:
                    file.write(value)

        :side effects: Given an optional image_directory, will write the images out
        to file.
        """
        images: dict[str, bytes] = {}
        for image in self.files_of_type("image"):
            with suppress(KeyError):
                images[os.path.basename(image.Target)] = self.zipf.read(image.path)
        if image_directory is not None:
            pathlib.Path(image_directory).mkdir(parents=True, exist_ok=True)
            for file, image_bytes in images.items():
                with open(os.path.join(image_directory, file), "wb") as image_copy:
                    _ = image_copy.write(image_bytes)
        return images


def _copy_but(
    in_zip: zipfile.ZipFile,
    out_zip: zipfile.ZipFile,
    exclusions: set[str] | None = None,
) -> None:
    """
    Copy every file in a docx except those listed in exclusions.

    :param in_zip: zipfile of origin xml file
    :param out_zip: zipfile of destination xml file
    :param exclusions: filenames you don't want to copy (e.g., 'document.xml')
    """
    exclusions = exclusions or set()
    for item in in_zip.infolist():
        if item.filename not in exclusions:
            buffer = in_zip.read(item.filename)
            out_zip.writestr(item, buffer)



"""Top-level code

:author: Shay Hill
:created: 7/2/2019
"""
from __future__ import annotations

from io import BytesIO
from pathlib import Path
from warnings import warn

from .docx_output import DocxContent
from .docx_reader import DocxReader


def docx2python(
    docx_filename: str | Path | BytesIO,
    image_folder: str | None = None,
    html: bool = False,
    paragraph_styles: bool = False,
    extract_image: bool | None = None,
    duplicate_merged_cells: bool = False,
) -> DocxContent:
    """
    Unzip a docx file and extract contents.

    :param docx_filename: path to a docx file
    :param image_folder: optionally specify an image folder
        (images in docx will be copied to this folder)
    :param html: bool, extract some formatting as html
    :param paragraph_styles: prepend the paragraphs style (if any, else "") to each
        paragraph. This will only be useful with ``*_runs`` attributes.
    :param extract_image: bool, extract images from document (default True)
    :param duplicate_merged_cells: bool, duplicate merged cells to return a mxn
        nested list for each table (default False)
    :return: DocxContent object
    """
    if extract_image is not None:
        warn(
            "'extract_image' is no longer a valid argument for docx2python. If an "
            + "image_folder is given as an argument to docx2python, images will be "
            + "written to that folder. A folder can be provided later with "
            + "``docx2python(filename).write_images(image_folder)``. Images files are "
            + "available as before with ``docx2text(filename).images`` attribute."
        )
    docx_context = DocxReader(
        docx_filename, html, paragraph_styles, duplicate_merged_cells
    )
    docx_content = DocxContent(docx_context, locals())
    if image_folder:
        _ = docx_content.images
    return docx_content



"""Extract text from docx content files.

:author: Shay Hill
:created: 6/6/2019

Content in the extracted docx is found in the ``word`` folder:
    ``word/document.html``
    ``word/header1.html``
    ``word/footer1.html``
"""
from __future__ import annotations

from contextlib import suppress
from typing import TYPE_CHECKING, List, Sequence, cast

from lxml.etree import _Element as EtreeElement  # type: ignore

from .attribute_register import Tags
from .bullets_and_numbering import BulletGenerator
from .depth_collector import DepthCollector, Run
from .forms import get_checkBox_entry, get_ddList_entry
from .iterators import iter_at_depth
from .namespace import qn
from .text_runs import (
    gather_Pr,
    get_paragraph_formatting,
    get_pStyle,
    get_run_formatting,
)

if TYPE_CHECKING:
    from docx_reader import File

TablesList = List[List[List[List[str]]]]


def _get_elem_depth(tree: EtreeElement) -> int | None:
    """What depth is this element in a nested list, relative to paragraphs (depth 4)?

    :param tree: element in a docx content xml (header, footer, officeDocument, etc.)

    :return: 4 - recursion depth;
        None if no paragraphs are found or if descending into nest would cause a
        false start (e.g., Tags.DOCUMENT or Tags.BODY which often have A paragraph (but
        not the next paragraph) at one or two levels down.

    Typically, the docx is a table of tables::

        [  # entire document
            [  # table
                [  # table row
                    [  # table cell
                        [  # paragraph
                            "",  # run
                            "",  # run
                            "",  # run
                        ]
                    ]
                ]
            ]
        ]

    But this isn't always the case. Instead of looking explicitly for tables,
    table rows, and table cells, look inside elements for paragraphs to determine
    depth in the nested list.

    E.g., given a table row element with a paragraph two levels in, return 2.
    So, depth of element will be 4 - 2 = 3.

    document = depth 0
    table = depth 1
    table row = depth 2
    table cell = depth 3
    paragraph = depth 4
    below paragraph = depth 5

    There will only ever be one document list, so the min depth returned is 1
    """

    if tree.tag in {Tags.DOCUMENT, Tags.BODY}:
        return None

    def search_at_depth(tree_: Sequence[EtreeElement], _depth: int = 0) -> int | None:
        """Width-first recursive search for Tags.PARAGRAPH

        :param tree_: a sequence of elements which may contain a paragraph
        :return: depth of the first paragraph found, or None if no paragraph found
        """
        if not tree_:
            return None
        if any(x.tag == Tags.PARAGRAPH for x in tree_):
            return max(4 - _depth, 1)
        return search_at_depth(sum([list(x) for x in tree_], []), _depth + 1)

    return search_at_depth([tree])


def get_paragraphs(file: File, root: EtreeElement) -> list[str]:
    """Return a list of paragraphs from the document

    :param file: an internal file element (e.g., header, footer, document))
    :param root: the root element of the document
    :return: a list of paragraphs
    """
    all_paragraphs: list[str] = []
    for branch in root:
        all_paragraphs += list(iter_at_depth(get_text(file, branch), 5))
    return all_paragraphs


def merged_text_tree(file: File, root: EtreeElement) -> str:
    """Return a string of all text in the document

    :param file: an internal file element (e.g., header, footer, document))
    :param root: the root element of the document
    :return: a string of all text in the document
    """
    return "".join(get_paragraphs(file, root))


def get_text(file: File, root: EtreeElement | None = None) -> TablesList:
    """Xml as a string to a list of cell strings.

    :param file: File instance from which text will be extracted.
    :param root: Optionally extract content from a single element.
        If None, root_element of file will be used.
    :return: A 5-deep nested list of strings.

    Sorts the text into the DepthCollector instance, five-levels deep

    ``[table][row][cell][paragraph][run]`` is a string

    Joins the runs before returning, so return list will be

    ``[table][row][cell][paragraph]`` is a string

    If you'd like to extend or edit this package, this function is probably where you
    want to do it. Nothing tricky here except keeping track of the text formatting.
    """
    root = root if root is not None else file.root_element
    bullets = BulletGenerator(file.context.numId2numFmts)
    # numId2count = _new_list_counter()
    tables = DepthCollector(5)

    xml2html = file.context.xml2html_format

    def branches(tree: EtreeElement) -> None:
        """
        Recursively iterate over tree. Add text when found.

        :param tree: An Element from an xml file (etree)
        :effect: Adds text cells to outer variable `tables`.
        """
        do_descend = True

        tree_depth = _get_elem_depth(tree)
        tables.set_caret(tree_depth)

        # queue up tags before opening any paragraphs or runs
        if tree.tag == Tags.PARAGRAPH:
            par = tables.commence_paragraph(get_paragraph_formatting(tree, xml2html))
            if file.context.do_pStyle:
                par.runs.insert(0, Run([], get_pStyle(tree) or "None"))
            tables.insert_text_as_new_run(bullets.get_bullet(tree))

        elif tree.tag == Tags.RUN:
            tables.commence_run(get_run_formatting(tree, xml2html))

        elif tree.tag in {Tags.TEXT, Tags.TEXT_MATH}:
            # oddly enough, these don't all contain text
            text = tree.text if tree.text is not None else ""
            if xml2html:
                text = text.replace("&", "&amp;")
                text = text.replace("<", "&lt;")
                text = text.replace(">", "&gt;")
            tables.add_text_into_open_run(text)

        elif tree.tag == Tags.MATH:
            # read equations
            text = "".join(str(x) for x in tree.itertext())
            do_descend = False
            tables.insert_text_as_new_run(f"<latex>{text}</latex>")

        elif tree.tag == Tags.BR:
            tables.add_text_into_open_run("\n")

        elif tree.tag == Tags.SYM:
            font = str(tree.attrib.get(qn("w:font")))
            char = str(tree.attrib.get(qn("w:char")))
            if char:
                tables.add_text_into_open_run(
                    f"<span style=font-family:{font}>&#x0{char[1:]};</span>"
                )

        elif tree.tag == Tags.FOOTNOTE:
            footnote_type = str(tree.attrib.get(qn("w:type"), "")).lower()
            if "separator" not in footnote_type:
                tables.insert_text_as_new_run(
                    f"footnote{str(tree.attrib[qn('w:id')])})\t"
                )

        elif tree.tag == Tags.ENDNOTE:
            endnote_type = str(tree.attrib.get(qn("w:type"), "")).lower()
            if "separator" not in endnote_type:
                tables.insert_text_as_new_run(
                    f"endnote{str(tree.attrib[qn('w:id')])})\t"
                )

        elif tree.tag == Tags.HYPERLINK:
            # look for an href, ignore internal references (anchors)
            text = merged_text_tree(file, tree)
            do_descend = False
            try:
                rId = tree.attrib[qn("r:id")]
                link = file.rels[rId]
                tables.insert_text_as_new_run(f'<a href="{link}">{text}</a>')
            except KeyError:
                tables.insert_text_as_new_run(text)

        if tree.tag == Tags.FORM_CHECKBOX:
            tables.insert_text_as_new_run(get_checkBox_entry(tree))

        elif tree.tag == Tags.FORM_DDLIST:
            tables.insert_text_as_new_run(get_ddList_entry(tree))

        elif tree.tag == Tags.FOOTNOTE_REFERENCE:
            tables.insert_text_as_new_run(
                f"----footnote{str(tree.attrib[qn('w:id')])}----"
            )

        elif tree.tag == Tags.ENDNOTE_REFERENCE:
            tables.insert_text_as_new_run(
                f"----endnote{str(tree.attrib[qn('w:id')])}----"
            )

        elif tree.tag == Tags.IMAGE:
            with suppress(KeyError):
                rId = tree.attrib[qn("r:embed")]
                image = file.rels[rId]
                tables.insert_text_as_new_run(f"----{image}----")

        elif tree.tag == Tags.IMAGEDATA:
            with suppress(KeyError):
                rId = tree.attrib[qn("r:id")]
                image = file.rels[rId]
                tables.insert_text_as_new_run(f"----{image}----")

        elif tree.tag == Tags.TAB:
            tables.insert_text_as_new_run("\t")

        if do_descend:
            for branch in tree:
                branches(branch)

        if tree.tag == Tags.PARAGRAPH:
            tables.conclude_paragraph()

        elif tree.tag == Tags.TABLE_CELL and file.context.duplicate_merged_cells:
            pr = gather_Pr(tree)

            if pr.get("vMerge", "Not None") is None:
                tables.set_caret(tree_depth)
                cell_idx = len(tables.caret) - 1
                assert isinstance(tree_depth, int)
                prev_row_cell = tables.view_branch((tree_depth - 2, -2, cell_idx))
                tables.caret[-1] = prev_row_cell

            grid_span = pr.get("gridSpan", 1)
            assert grid_span is not None
            for _ in range(int(grid_span) - 1):
                tables.set_caret(tree_depth)
                tables.caret.append(tables.caret[-1])

        elif tree.tag == Tags.RUN:
            tables.conclude_run()

        tables.set_caret(tree_depth)

    branches(root)

    if tables.orphan_runs:
        _ = tables.commence_paragraph()
    if tables.open_pars:
        tables.conclude_paragraph()

    return cast(TablesList, tables.tree)



""" Content from files that aren't ``word/document.xml``

:author: Shay Hill
:created: 6/26/2019

Most of the "meat" in a docx file is in ``word/document.xml``. These functions retrieve
numbering formats, images, and font styles from *other* files in a decompressed docx.
"""
from __future__ import annotations

import re
import zipfile

from lxml import etree
from lxml.etree import _Element as EtreeElement  # type: ignore

from .namespace import qn


def collect_numFmts(numFmts_root: EtreeElement) -> dict[str, list[str]]:
    """
    Collect abstractNum bullet formats into a dictionary

    :param numFmts_root: Root element of ``word/numbering.xml``.
    :return: numId mapped to numFmts (by ilvl)

    :background:

    ``word/numbering.xml`` will have two sections.

    **SECTION 1** - Some abstractNum elements defining numbering formats for multiple
    indentation levels::

        <w:abstractNum w:abstractNumId="0">
            <w:lvl w:ilvl="0"><w:numFmt w:val="decimal"/></w:lvl>
            <w:lvl w:ilvl="1"><w:numFmt w:val="lowerLetter"/></w:lvl>
            ...
        </w:abstractNum>

    **SECTION 2** - Some num elements, each referencing an abstractNum. Multiple nums
    may reference the same abstractNum, but each will maintain a separate count (i.e.,
    each numbered paragraph will start from 1, even if it shares a style with another
    paragraph.)::

        <w:num w:numId="1">
            <w:abstractNumId w:val="0"/>
        </w:num>
        <w:num w:numId="2">
            <w:abstractNumId w:val="5"/>
        </w:num>

    **E.g, Given**: *above*

    **E.g., Returns**::

        {
            # -----ilvl=0------ilvl=1------ilvl=2---
            "1": ["decimal", "lowerLetter", ...],
            "2": ...
        }
    """
    abstractNumId2numFmts: dict[str, list[str]] = {}

    for abstractNum in numFmts_root.findall(qn("w:abstractNum")):
        id_ = str(abstractNum.attrib[qn("w:abstractNumId")])
        abstractNumId2numFmts[id_] = []
        for lvl in abstractNum.findall(qn("w:lvl")):
            numFmt = lvl.find(qn("w:numFmt"))
            if numFmt is not None:
                abstractNumId2numFmts[id_].append(str(numFmt.attrib[qn("w:val")]))

    numId2numFmts: dict[str, list[str]] = {}
    num: EtreeElement
    for num in (x for x in numFmts_root.findall(qn("w:num")) if x is not None):
        numId = num.attrib[qn("w:numId")]
        if numId is None:
            continue
        abstractNumId = num.find(qn("w:abstractNumId"))
        if abstractNumId is None:
            continue
        abstractNumIdval = abstractNumId.attrib.get(qn("w:val"))
        numId2numFmts[str(numId)] = abstractNumId2numFmts[str(abstractNumIdval)]

    return numId2numFmts


def collect_rels(zipf: zipfile.ZipFile) -> dict[str, list[dict[str, str]]]:
    """
    Map file to relId to attrib

    :param zipf: created by ``zipfile.ZipFile("docx_filename")``
    :return: a deep dictionary ``{filename: list of Relationships``

    Each rel in list of Relationships is::

        {
            "Id": "rId1",
            "Type": "http...",
            "Target": "path to file in docx"
        }

    There are several rels files:

    ``_rels/.rels``: rels related to entire structure.  The identity of
        ``word/document.xml`` is here. (It might be called ``word/document2.xml`` or
        something else. Checking here is the best way to make sure.)

    ``word/_rels/document.xml.rels``: images, headers, etc. referenced by
        ``word/document.xml``

    ``word/_rels/header1.xml.rels``: images, etc. for ``header1.xml``

    ...

    Get everything from everywhere. Map ``_rels/.rels`` to ``'rels'`` and everything
    else to e.g., ``'document'`` or ``'header'``. RelIds are **not** unique between
    these files.

    **E.g, Given**::

    # one of several files

        <?xml version="1.0" encoding="UTF-8" standalone="yes"?>
        <Relationships xmlns="http://schemas.../relationships">
            <Relationship Id="rId3" Type="http://schemas... \
                /extended-properties" Target="docProps/app.xml"/>
            <Relationship Id="rId2" Type="http://schemas... \
                /core-properties" Target="docProps/core.xml"/>
            <Relationship Id="rId1" Type="http://schemas... \
                /officeDocument" Target="word/document.xml"/>
            <Relationship Id="rId4" Type="http://schemas... \
                /custom-properties" Target="docProps/custom.xml"/>
        </Relationships>

    **Returns**::

        {
            "filename": [
                {
                    "Id": "rId3",
                    "Type": "http://schemas.../extended-properties",
                    "Target": "docProps/app.xml",
                },
                {
                    "Id": "rId2",
                    "Type": "http://schemas.../core-properties",
                    "Target": "docProps/core.xml",
                },
            ]
        }
    """
    path2rels: dict[str, list[dict[str, str]]] = {}
    for rels in (x for x in zipf.namelist() if x[-5:] == ".rels"):
        path2rels[rels] = [
            {str(y): str(z) for y, z in x.attrib.items()}
            for x in etree.fromstring(zipf.read(rels))
        ]
    return path2rels


def collect_docProps(root: EtreeElement) -> dict[str, str | None]:
    """
    Get author, modified, etc. from core-properties (should be docProps/core.xml)

    :param root: root of the XML tree
    :return: document property names mapped to values

    **E.g., Given**::

        <cp:coreProperties xmlns:cp="http://schemas.openxmlformats...">
            <dc:title>SG-DOP-5009 - Operate ROMAR swarf unit
            </dc:title>
            <dc:creator>Shay Hill
            </dc:creator>
            <cp:lastModifiedBy>Shay Hill
            </cp:lastModifiedBy>
            <cp:revision>6
            </cp:revision>
            <cp:lastPrinted>2017-11-17T15:47:00Z
            </cp:lastPrinted>
            <dcterms:created xsi:type="dcterms:W3CDTF">2019-01-10T07:21:00Z
            </dcterms:created>
            <dcterms:modified xsi:type="dcterms:W3CDTF">2019-01-11T11:41:00Z
            </dcterms:modified>
        </cp:coreProperties>

    **E.g., Returns**::

        {
            "title": "SG-DOP-5009 - Operate ROMAR swarf unit",
            "creator": "Shay Hill",
            "lastModifiedBy": "Shay Hill",
            "revision": "6",
            ...
        }
    """
    docProp2text: dict[str, str | None] = {}
    capture_tag_name = re.compile(r"{.+}(?P<tag_name>\w+)")
    for dc in root:
        tag_match = re.match(capture_tag_name, dc.tag)
        if tag_match:
            docProp2text[tag_match.group("tag_name")] = dc.text
    return docProp2text



"""Import function docx2python into the docx2python namespace.

:author: Shay Hill
:created: 2023-01-09
"""

from .main import docx2python

__all__ = ["docx2python"]

# TODO: remove import cycles (turn back on in pyproject.toml)



"""Collect xml text in a nested list

:author: Shay Hill
:created: 6/26/2019

::

    [  # tables
        [  # table
            [  # row
                [  # cell
                    [  # paragraph
                        ""  # text run
                    ]
                ]
            ]
        ]
    ]

Text in table cells and text outside of a table must be captured at depth=5. To keep
track of this, these few methods will put text where it needs to be.

The package will recursively descend into elements in the docx file, so the point at
which a table, row, cell, paragraph, or text run begins and ends is known. Drop and
raise the caret when these items are opened or closed in the xml. Insert text when
found.

Shorthand for this package. Instances of this class should not escape the package.
Pass out of package with depth_collector_instance.tree.
"""

from __future__ import annotations

from contextlib import suppress
from dataclasses import dataclass, field
from typing import Any, Iterable

from .text_runs import html_close, html_open


@dataclass
class Run:
    """A text run. Html styles and text content"""

    html_style: list[str] = field(default_factory=list)
    text: str = field(default="")

    def __str__(self) -> str:
        """Return any text content in the run

        :return: text content or "" if none
        """
        if self.text:
            return html_open(self.html_style) + self.text + html_close(self.html_style)
        return ""


@dataclass
class Par:
    """A text paragraph. Html styles and a list of run strings"""

    html_style: list[str]
    runs: list[Run] = field(default_factory=list)

    @property
    def strings(self) -> list[str]:
        """Return a list of strings from the runs

        :return: a string for each run with text content
        """
        return [x for x in (str(y) for y in self.runs) if x]


class CaretDepthError(Exception):
    """Caller attempted to raise or lower DepthCollector caret out of range"""


class DepthCollector:
    """Insert items into a tree at a consistent depth."""

    def __init__(self, item_depth: int) -> None:
        """
        Record item depth and initiate data container.

        :param item_depth: content will only appear at this depth, though empty lists
            may appear above. I.e., this is how many brackets to open before inserting
            an item. E.g., item_depth = 3 => [[['item']]].
        """
        # TODO: factor out item_depth
        self.item_depth = item_depth
        self._par_depth = 4

        self._rightmost_branches: list[Any] = [[]]

        self.open_pars: list[Par] = []
        self.orphan_runs: list[Run] = []

    def view_branch(self, address: Iterable[int]) -> Any:
        """Return the item at the given address

        :param address: a tuple of indices to the item to be returned
        :return: the item at the address
        """
        branch = self._rightmost_branches
        for i in address:
            branch = branch[i]
        return branch

    @staticmethod
    def _get_run_strings(runs: list[Run]) -> list[str]:
        """Return a string for each run in the current open paragraph. Ignore ""

        :param runs: list of runs
        :return: a string for each run with text content
        """
        return [x for x in (str(x) for x in runs) if x]

    def commence_paragraph(self, html_style: list[str] | None = None) -> Par:
        """Gather any cached runs and open a new paragraph.

        :param html_style: html style to apply to the paragraph
        :return: the new paragraph
        """
        html_style = html_style or []
        new_par = Par(html_style, self.orphan_runs + [Run([], html_open(html_style))])
        self.orphan_runs = []
        self.open_pars.append(new_par)
        return new_par

    def conclude_paragraph(self) -> None:
        """Close the current paragraph and add it to the tree."""
        old_par = self.open_pars.pop()
        old_par.runs.append(Run([], html_close(old_par.html_style)))
        self.insert(old_par.strings)

    def commence_run(self, html_style: list[str] | None = None) -> None:
        """Open a new run and add it to the current paragraph.

        :param html_style: html style to apply to the run
        """
        self._open_runs.append(Run(html_style or []))

    def conclude_run(self) -> None:
        """Close the current run and add it to the current paragraph."""
        self.commence_run()

    @property
    def tree(self) -> list[str | list[str]]:
        """All collected items.

        :return: a nested list of _par_depth + 1 levels
        """
        return self._rightmost_branches[0]

    @property
    def caret(self) -> list[str | list[str]]:
        """Lowest open child.

        :return: the list where new content will be appended
        """
        return self._rightmost_branches[-1]

    @property
    def caret_depth(self) -> int:
        """Depth of the lowest open child.

        :return: from 0 to _par_depth, the depth of the last-closed element in the
            tree.
        """
        return len(self._rightmost_branches)

    @property
    def _open_runs(self) -> list[Run]:
        """Runs in the current paragraph.

        :return: a list of runs
        """
        with suppress(IndexError):
            return self.open_pars[-1].runs
        return self.orphan_runs

    @property
    def _open_run(self) -> Run:
        """The last run in the current paragraph.

        :return: a run
        """
        if not self._open_runs:
            self._open_runs.append(Run())
        return self._open_runs[-1]

    def _drop_caret(self) -> None:
        """Create a new branch under caret.

        :raise CaretDepthError: if caret is already at the maximum depth
        :return: None
        """
        if self.caret_depth >= self.item_depth:
            raise CaretDepthError("will not lower caret beneath item_depth")
        self._rightmost_branches[-1].append([])
        self._rightmost_branches.append(self._rightmost_branches[-1][-1])

    def _raise_caret(self) -> None:
        """Close branch at caret and move up to parent.

        :raise CaretDepthError: if there is no outside list to which to ascend
        """
        if self.caret_depth == 1:
            raise CaretDepthError("will not raise caret above root")
        self._rightmost_branches = self._rightmost_branches[:-1]

    def set_caret(self, depth: None | int) -> None:
        """
        Set caret at given depth.

        :param depth: depth level for caret (between 1 and item_depth inclusive)
            another at the same depth. This is how consecutive paragraphs avoid being
            merged into one paragraph. You'll want this true for every element except
            text runs. :depth: == None means the element (perhaps ``body``) does not
            effect depth (see details in docx_text._get_elem_depth).
        """
        if depth is None:
            return
        while self.caret_depth < depth:
            self._drop_caret()
        while self.caret_depth > depth:
            self._raise_caret()

    def insert(self, item: list[str]) -> None:
        """Add item at self._par_depth. Add branches if necessary to reach depth.

        :param item: list of strings to insert at caret

        This dumps the contents of the most recently closed paragraph into the
        _rightmost_branches collector.
        """
        self.set_caret(self._par_depth)
        self._rightmost_branches[-1].append(item)

    def add_text_into_open_run(self, item: str) -> None:
        """
        Add item into previous run.

        :param item: string to insert into previous run

        This is for tags and other text that appears between run tags. All entries to
        ``add_text_into_open_run`` will be merged together.
        """
        self._open_run.text += item

    def insert_text_as_new_run(self, item: str, styled: bool = False) -> None:
        """
        Close previous run, cache style, open and close new run, re-open cached style.

        :param item: string to insert into new run
        :param styled: True if item has associated html styles

        This is for items like links that may be inside a run element with other text.

        Paraphrased in html:

            <run><b>some text<a href="">link</a>other text</b></run>

        Starts with an open run
            <run><b>some text

        Then hits the link. We'll make this a run inside the actual run

            <run><b>some text</b></run>  # close this open run
            <run><a href="">link</a></run>  # add link as a new run
            <run><b>  # open a new run with the same style as the aborted first run
        """
        open_style = self._open_run.html_style
        if styled:
            self._open_runs.append(Run(open_style, item))
        else:
            self._open_runs.append(Run([], item))
        self._open_runs.append(Run(open_style))



"""Register namespace entries in xml ``document`` elements.

:author: Shay Hill
:created: 7/5/2019

A ``<w:document>`` element at the top of each xml file defines a namespace::

    <w:document
        xmlns:r="http://schemas.openxmlformats.org/officeDocument/2006/relationships"
        xmlns:w="http://schemas.openxmlformats.org/wordprocessingml/2006/main"
    />

These entries can be accessed in the file by their abbreviations::

    <w:p>
        contents of paragraph
    </w:p>

``lxml.etree`` reads ``"<w:p>"`` as

``"{http://schemas.openxmlformats.org/wordprocessingml/2006/main}p"``

This module defines the necessary namespaces and transforms ``"w:p"`` to
``{http://...}p``. This allows readable code like::

    if element.tag == qn("w:p"):

instead of::

    if element.tag == "{http://schemas.openxmlformats.org/wordprocessingml/2006/main}p":

If somewhere along the line this package just stops working, it may be that the NSMAP
entries have been updated for whatever docx you're working with (though that's not
supposed to ever happen). *If* this happens::

    1) Unzip the docx.
    2) open ``word/document.xml`` in a text editor.
    3) Search for xmlns:w=[some string]
    4) update NSMAP['w'] = some string

Lxml allows (deceptively) easy access to a file's namespaces; however, this is
problematic because ``root_element.nsmap`` may not retrieve all nsmap entries. Other
entries may be buried inside sub-environments further down in the tree. It is safer
to explicate namespace mapping.

If you extend docx2text with other tags, additional NSMAP entries may be necessary.
"""


NSMAP = {
    "a": "http://schemas.openxmlformats.org/drawingml/2006/main",
    "cp": "http://schemas.openxmlformats.org/package/2006/metadata/core-properties",
    "dc": "http://purl.org/dc/elements/1.1/",
    "dcterms": "http://purl.org/dc/terms/",
    "m": "http://schemas.openxmlformats.org/officeDocument/2006/math",
    "r": "http://schemas.openxmlformats.org/officeDocument/2006/relationships",
    "v": "urn:schemas-microsoft-com:vml",
    "w": "http://schemas.openxmlformats.org/wordprocessingml/2006/main",
}


def qn(tag: str) -> str:
    """
    Turn a namespace-prefixed tag into a Clark-notation qualified tag.

    :param tag: namespace-prefixed tag, e.g. ``w:p``
    :return: Clark-notation qualified tag,
        e.g. ``{http://schemas.openxmlformats.org/wordprocessingml/2006/main}p``

    Stands for 'qualified name', a utility function to turn a namespace prefixed tag
    name into a Clark-notation qualified tag name for lxml.

        >>> qn('w:cSld')
        '{http://schemas.../main}cSld'

    Source: https://github.com/python-openxml/python-docx/
    """
    prefix, tagroot = tag.split(":")
    uri = NSMAP[prefix]
    return f"{{{uri}}}{tagroot}"



"""Get text run formatting.

:author: Shay Hill
:created: 7/4/2019

Text runs are formatted inline in the ``trash/document.xml`` or header files. Read
those elements to extract formatting information.
"""
from __future__ import annotations

import re
from collections import defaultdict
from contextlib import suppress
from typing import Sequence, Tuple, Union

from lxml.etree import _Element as EtreeElement  # type: ignore

from .attribute_register import HtmlFormatter, Tags
from .namespace import qn


def _elem_tag_str(elem: EtreeElement) -> str:
    """The text part of an elem.tag (the portion right of the colon)

    :param elem: an xml element
    :return: the text part of an elem.tag (the portion right of the colon)
    :raise RuntimeError: if the tag is not qualified

    create with:

        document = etree.fromstring('bytes string')
        # recursively search document elements.

    **E.g., given**:

        document = etree.fromstring('bytes string')
        # document.tag = '{http://schemas.openxml.../2006/main}:document'
        elem_tag_str(document)

    **E.g., returns**:

        'document'
    """
    tag_str = re.match(r"{.*}(?P<tag_name>\w+)", elem.tag)
    if tag_str:
        return tag_str.group("tag_name")
    raise RuntimeError("could not get tag from elem")


def _gather_sub_vals(element: EtreeElement, qname: str) -> dict[str, str | None]:
    """Gather formatting elements for a paragraph or text run.

    :param element: a ``<w:r>`` or ``<w:p>`` xml element. Maybe others
    :param qname: qualified name for child element.

    create with::

        document = etree.fromstring('bytes string')
        # recursively search document for <w:r> elements.

    :return: Style names ('b/', 'sz', etc.) mapped to values.

    To keep things more homogeneous, I've given tags like ``<w:b/>`` (bold) a value of
    None, even though they don't take a value in xml.

    Each element of rPr will be either present (returned tag: None) or have a value
    (returned tag: val).

    **E.g., given**::

         <w:r w:rsidRPr="000E1B98">
            <w:rPr>
                <w:rFonts w:ascii="Arial"/>
                <w:b/>
                <w:sz w:val="32"/>
                <w:szCs w:val="32"/>
                <w:u w:val="single"/>
            </w:rPr>
            <w:t>text styled  with rPaa
            </w:t>
        </w:r>

    **E.g., returns**::

        {
            "rFonts": True,
            "b": None,
            "u": "single",
            "i": None,
            "sz": "32",
            "color": "red",
            "szCs": "32",
        }
    """

    sub_vals: dict[str, str | None] = {}
    with suppress(StopIteration):
        for sub_element in next(element.iterfind(qname)):
            sub_val = sub_element.attrib.get(qn("w:val"))
            if sub_val:
                sub_vals[_elem_tag_str(sub_element)] = str(sub_val)
            else:
                sub_vals[_elem_tag_str(sub_element)] = None
    return sub_vals


def gather_Pr(element: EtreeElement) -> dict[str, str | None]:
    """
    Gather style values for a <w:r>, <w:tc>, or <w:p> element (maybe others)

    :param element: any xml element. r and p elems typically have Pr values.
    :return: Style names ('b/', 'sz', etc.) mapped to values.

    These elements often have a subelement ``<w:pPr>`` or ``<w:rPr>`` which contains
    formatting instructions. This includes colspan, rowspan, and other table-cell
    properties.

    Will infer a style element qualified name: p -> pPr; r -> rPr

    Call this with any element. Runs and Paragraphs may have a Pr element. Most
    elements will not, but the function will will quietly return an empty dict.
    """
    return _gather_sub_vals(element, element.tag + "Pr")


def get_pStyle(paragraph_element: EtreeElement) -> str:
    """
    Collect and format paragraph -> pPr -> pStyle value.

    :param paragraph_element: a ``<w:p>`` xml element

    :return: ``[(pStyle value, '')]``

    Also see docstring for ``gather_pPr``
    """
    return gather_Pr(paragraph_element).get("pStyle", "") or ""


def get_run_formatting(
    run_element: EtreeElement, xml2html: dict[str, HtmlFormatter]
) -> list[str]:
    """
    Get run-element formatting converted into html.

    :param run_element: a ``<w:r>`` xml element
        create with::

            document = etree.fromstring('bytes string')
            # recursively search document for <w:r> elements.

    :param xml2html: mapping to convert xml styles to html styles
        e.g., {
            'b': (<function <lambda> at 0x0000026BC7875A60>,),
            'smallCaps': (<function <lambda> at 0x0000026BC7896DC0>, 'font', 'style')
        }

    :return: ``['b', 'i', ...]``

    Lists are always returned in order:

    ``"span"`` first then any other styles in alphabetical order.

    Also see docstring for ``gather_rPr``
    """
    return _format_Pr_into_html(gather_Pr(run_element), xml2html)


def get_paragraph_formatting(
    paragraph_element: EtreeElement, xml2html: dict[str, HtmlFormatter]
) -> list[str]:
    """
    Get paragraph-element formatting converted into html.

    :param paragraph_element: a ``<w:p>`` xml element
        create with::

            document = etree.fromstring('bytes string')
            # recursively search document for <w:r> elements.

    :param xml2html: mapping to convert xml styles to html styles
        e.g., {
            'b': (<function <lambda> at 0x0000026BC7875A60>,),
            'smallCaps': (<function <lambda> at 0x0000026BC7896DC0>, 'font', 'style')
        }

    :return: ``['b', 'i', ...]``

    Tuples are always returned in order:

    ``"font"`` first then any other styles in alphabetical order.

    Also see docstring for ``gather_rPr``
    """
    return _format_Pr_into_html({get_pStyle(paragraph_element): None}, xml2html)


def _format_Pr_into_html(
    Pr2val: dict[str, str | None], xml2html: dict[str, HtmlFormatter]
) -> list[str]:
    """
    Format tags and values into html strings.

    :param Pr2val: tags mapped to values (extracted from xml)
        e.g., {'b': None, 'bCs': None}
    :param xml2html: mapping to convert xml styles to html styles
        e.g., {
            'b': (<function <lambda> at 0x0000026BC7875A60>,),
            'smallCaps': (<function <lambda> at 0x0000026BC7896DC0>, 'span', 'style')
        }
    :return: the interior part of html opening tags, eg, ['span style="..."', 'b', 'i']

    Types of styles supported:
    (None, None, formatter -> tag, None)
        -> outside any containers, no value set, e.g., `<b>`
    ('span', 'style', formatter -> tag, val)
        -> inside a span, inside a style property, e.g., `<span style="tag: val">`

    Other formats would probably work, but they aren't necessary to support the tags
    supported (see README).
    """
    style: list[str] = []

    # group together supported formats with the same container and property_
    # e.g., group together everything that goes into `<span style="$HERE$">`
    # con_pro2for[(con, pro)] = string created from for
    cp2f_keytype = Tuple[Union[None, str], Union[None, str]]
    con_pro2for: defaultdict[cp2f_keytype, list[str]] = defaultdict(list)
    for tag, val in ((k, v) for k, v in Pr2val.items() if k in xml2html):
        formatter, container, property_ = xml2html[tag]
        con_pro2for[(container, property_)].append(formatter(tag, val or ""))

    # group together supported formats with the same container
    # e.g., group together everything that goes into `<span $HERE$>`
    # con2pro_for[(con,)] = string created from pro and for
    con2pro_for: defaultdict[str, list[str]] = defaultdict(list)
    for k, v in sorted((k, v) for k, v in con_pro2for.items() if k[1] is not None):
        con2pro_for[k[0] or ""].append(f'{k[1]}="{";".join(sorted(v))}"')

    # incorporate container type into string
    # style.append(string created from con, pro, and for)
    for k_, v_ in sorted((k, v) for k, v in con2pro_for.items() if k):
        style.append(f"{k_} {' '.join(v_)}")

    # add back in formats with no container or property_
    style += sorted(con_pro2for[(None, None)])
    return style


def get_html_formatting(
    elem: EtreeElement, xml2html: dict[str, HtmlFormatter]
) -> list[str]:
    """
    Get style for an element (if available)

    :param elem: a run or paragraph element.
    :param xml2html: mapping to convert xml styles to html styles
        e.g., {
            'b': (<function <lambda> at 0x0000026BC7875A60>,),
            'smallCaps': (<function <lambda> at 0x0000026BC7896DC0>, 'font', 'style')
        }
    :return: ``[(rPr, val), (rPr, val) ...]``
    """
    if elem.tag == Tags.RUN:
        return get_run_formatting(elem, xml2html)
    if elem.tag == Tags.PARAGRAPH:
        return get_paragraph_formatting(elem, xml2html)
    return []


def html_open(style: Sequence[str]) -> str:
    """
    HTML tags to open a style.

    :param style: sequence of html tags without the '<' and '>'
    :return: opening html tags joined into a single string

    >>> style = ['font color="red" size="32"', 'b', 'i', 'u']
    >>> html_open(style)
    '<font color="red" size="32"><b><i><u>'
    """
    return "".join(f"<{x}>" for x in style)


def html_close(style: list[str]) -> str:
    """
    HTML tags to close a style.

    :param style: sequence of html tags without the '<' and '>'
    :return: closing html tags joined into a single string

    >>> style = ['font color="red" size="32"', 'b', 'i', 'u']
    >>> html_close(style)
    '</u></i></b></font>'

    Tags will always be in reverse (of open) order, so open - close will look like::

        <b><i><u>text</u></i></b>
    """
    return "".join(f"</{x.split()[0]}>" for x in reversed(style))



"""Output format for extracted docx content.

:author: Shay Hill
:created: 7/5/2019

Holds runs in a 5-deep nested list (paragraphs are lists of text runs [strings])::

    [  # tables
        [  # table
            [  # row
                [  # cell
                    [  # paragraph
                        "run 1 ",  # text run
                        "run 2 ",  # text run
                        "run 3"  # text run
                    ]
                ]
            ]
        ]
    ]

_runs properties (e.g., ``header_runs``) return text in this format.

Also returns a 4-deep nested list (paragraphs are strings)::

    [  # tables
        [  # table
            [  # row
                [  # cell
                    "run 1 run 2 run 3"  # paragraph
                ]
            ]
        ]
    ]

This is the format for default (no trailing "_runs", e.g ``header``) properties.

"""
from copy import deepcopy
from dataclasses import dataclass
from typing import Any, Dict, Optional
from warnings import warn

from .docx_context import collect_docProps
from .docx_reader import DocxReader
from .docx_text import TablesList
from .iterators import enum_at_depth, get_html_map, iter_at_depth


@dataclass
class DocxContent:
    """Holds return values for docx content."""

    docx_reader: DocxReader
    docx2python_kwargs: Dict[str, Any]

    def close(self):
        """Close the zipfile opened by DocxReader"""
        self.docx_reader.close()

    def __enter__(self) -> "DocxContent":
        """Do nothing. The zipfile will open itself when needed.

        :return: self"""
        return self

    def __exit__(
        self,
        exc_type: Any,  # None | Type[Exception], but py <= 3.9 doesn't like it.
        exc_value: Any,  # None | Exception, but py <= 3.9 doesn't like it.
        exc_traceback: Any,  # None | TracebackType, but py <= 3.9 doesn't like it.
    ):
        """Close the zipfile opened by DocxReader

        :param exc_type: Python internal use
        :param exc_value: Python internal use
        :param exc_traceback: Python internal use
        """
        self.close()

    def __getattr__(self, name: str) -> Any:
        """
        Create depth-four paragraph tables form depth-five run tables.

        :param name: name of an internal docx xml file
        :return: extracted text from named file with runs joined together into
            paragraphs.
        :raise AttributeError: if "name" file cannot be found

        Docx2Python v1 joined runs into paragraphs [[[str]]] earlier in the code.
        Docx2Python v2 exposes runs [[[[str]]]] to the user, but still returns
        paragraphs by default.
        """
        if name in {"header", "footer", "body", "footnotes", "endnotes"}:
            runs = deepcopy(getattr(self, name + "_runs"))
            for (i, j, k, l), paragraph in enum_at_depth(runs, 4):
                runs[i][j][k][l] = "".join(paragraph)
            return runs
        raise AttributeError(f"no attribute {name}")

    def _get_runs(self, type_: str) -> TablesList:
        """Get text runs for an internal document type.

        :param type_: this package looks for any of
            ("header", "officeDocument", "footer", "footnotes", "endnotes")
            You can try others.
        :return: text runs [[[[str]]]]
        """
        content: TablesList = []
        for file in self.docx_reader.files_of_type(type_):
            content += file.content
        return content

    @property
    def header_runs(self) -> TablesList:
        """Get text runs for header files.

        :return: text runs [[[[str]]]]
        """
        return self._get_runs("header")

    @property
    def footer_runs(self) -> TablesList:
        """Get text runs for footer files.

        :return: text runs [[[[str]]]]
        """
        return self._get_runs("footer")

    @property
    def officeDocument_runs(self) -> TablesList:
        """Get text runs for the main officeDocument file.

        :return: text runs [[[[str]]]]
        """
        return self._get_runs("officeDocument")

    @property
    def body_runs(self) -> TablesList:
        """Get text runs for the main officeDocument file.

        :return: text runs [[[[str]]]]

        This is an alias for officeDocument_runs.
        """
        return self.officeDocument_runs

    @property
    def footnotes_runs(self) -> TablesList:
        """Get text runs for footnotes files.

        :return: text runs [[[[str]]]]
        """
        return self._get_runs("footnotes")

    @property
    def endnotes_runs(self) -> TablesList:
        """Get text runs for endnotes files.

        :return: text runs [[[[str]]]]
        """
        return self._get_runs("endnotes")

    @property
    def images(self) -> Dict[str, bytes]:
        """Get bytestring of all images in the document.

        :return: dict {image_name: image_bytes}
        """
        return self.docx_reader.pull_image_files(
            self.docx2python_kwargs["image_folder"]
        )

    @property
    def document(self) -> TablesList:
        """All docx "tables" concatenated.

        :return: text paragraphs [[[str]]]
        """
        return self.header + self.body + self.footer + self.footnotes + self.endnotes

    @property
    def document_runs(self) -> TablesList:
        """All docx x_runs properties concatenated.

        :return: text runs [[[[str]]]]
        """
        return (
            self.header_runs
            + self.body_runs
            + self.footer_runs
            + self.footnotes_runs
            + self.endnotes_runs
        )

    @property
    def text(self) -> str:
        """All docx paragraphs, "\n\n" joined.

        :return: all docx paragraphs, "\n\n" joined
        """
        if self.docx2python_kwargs["paragraph_styles"] is True:
            # Paragraph descriptors have been inserted as the first run of each
            # paragraph. Take them out.
            pars = ["".join(x[1:]) for x in iter_at_depth(self.document_runs, 4)]
            return "\n\n".join(pars)
        return "\n\n".join(iter_at_depth(self.document, 4))

    @property
    def html_map(self) -> str:
        """A visual mapping of docx content.

        :return: html to show all strings with index tuples
        """
        return get_html_map(self.document)

    @property
    def properties(self) -> Dict[str, Optional[str]]:
        """Document core-properties as a dictionary.

        :return: document core-properties as a dictionary

        Docx files created with Google docs won't have core-properties. If the file
        `core-properties` is missing, return an empty dict.
        """
        warn(
            "DocxContent.properties is deprecated and will be removed in some future "
            + "version. Use DocxContent.core_properties.",
            FutureWarning,
        )
        return self.core_properties

    @property
    def core_properties(self) -> Dict[str, Optional[str]]:
        """Document core-properties as a dictionary.

        :return: document core-properties as a dictionary

        Docx files created with Google docs won't have core-properties. If the file
        `core-properties` is missing, return an empty dict.
        """
        try:
            docProps = next(iter(self.docx_reader.files_of_type("core-properties")))
            return collect_docProps(docProps.root_element)
        except StopIteration:
            warn(
                "Could not find core-properties file (should be in docProps/core.xml) "
                + "in DOCX, so returning an empty core_properties dictionary. Docx "
                + "files created in Google Docs do not have a core-properties file, "
                + "so this may be expected."
            )
            return {}

    def save_images(self, image_folder: str) -> Dict[str, bytes]:
        """Write images to hard drive.

        :param image_folder: folder to write images to
        :return: dictionary of image names and image data

        If the image folder does not exist, it will not be created.
        """
        return self.docx_reader.pull_image_files(image_folder)



""" Generate bullet and numbered-list strings.

:author: Shay Hill
:created: 11/15/2021

Docx xml files do not track explicit numbering values. Each numbered paragraph has ::

    <w:ilvl w:val="0"/>   # indentation level
    <w:numId w:val="9"/>  # index to a list [by ilvl] of numbered-list formats

Docx2Python keeps track of current numbering value, and increments these values as
numbered paragraphs are encountered. If extracting partial text, the numbers may be
incorrect, because all paragraphs in a numbered-list format may not be encountered
during the extraction.
"""
from __future__ import annotations

import warnings
from collections import defaultdict
from typing import Callable

from lxml.etree import _Element as EtreeElement  # type: ignore

from docx2python import numbering_formats as nums
from docx2python.namespace import qn


def _get_bullet_function(numFmt: str) -> Callable[[int], str]:
    """Select a bullet or numbering format function from xml numFmt.

    :param numFmt: xml numFmt (e.g., decimal, lowerLetter)
    :return: a function that takes an int and returns a string. If numFmt is not
        recognized, treat numbers as bullets.
    """
    numFmt2bullet_function: dict[str, Callable[[int], str]] = {
        "decimal": nums.decimal,
        "lowerLetter": nums.lower_letter,
        "upperLetter": nums.upper_letter,
        "lowerRoman": nums.lower_roman,
        "upperRoman": nums.upper_roman,
        "bullet": nums.bullet,
    }
    try:
        retval_: Callable[[int], str] = numFmt2bullet_function[numFmt]
        return retval_
    except KeyError:
        warnings.warn(
            f"{numFmt} numbering format not implemented, "
            + f"substituting '{nums.bullet()}'"
        )
        return nums.bullet


def _new_list_counter() -> defaultdict[str, defaultdict[str, int]]:
    """
    A counter, starting at zero, for each numId

    :return: {
        a_numId: 0,
        b_numId: 0
    }

    This is what you need to keep track of where every nested list is at.
    """
    return defaultdict(lambda: defaultdict(lambda: 0))


def _increment_list_counter(ilvl2count: defaultdict[str, int], ilvl: str) -> int:
    """
    Increase counter at ilvl, reset counter at deeper levels.

    :param ilvl2count: context['numId2count']
    :param ilvl: string representing an integer
    :return: updated count at ilvl.
        updates context['numId2count'] by reference

    On a numbered list, the count for sub-lists should reset when a parent list
    increases, e.g.,

    1. top-level list
        a. sublist
        b. sublist continues
    2. back to top-level list
        a. sublist counter has been reset

    List counters are defaultdicts, so we can reset sublist counters by deleting
    them.
    """
    ilvl2count[ilvl] += 1
    deeper_levels = [x for x in ilvl2count.keys() if x > ilvl]
    for level in deeper_levels:
        del ilvl2count[level]
    return ilvl2count[ilvl]


class BulletGenerator:
    """
    Keep track of list counters and generate bullet strings.
    """

    def __init__(self, numId2numFmts: dict[str, list[str]]) -> None:
        """
        Set numId2numFmts. Initiate counters.
        """
        self.numId2numFmts = numId2numFmts
        self.numId2count = _new_list_counter()

    def get_bullet(self, paragraph: EtreeElement) -> str:
        """
        Get bullet string if paragraph is numbered. (e.g, '--  ' or '1)  ')

        :param paragraph: <w:p> xml element
        :return: specified 'bullet' string or '' if paragraph is not numbered

        <w:p>
            <w:pPr>
                <w:numPr>
                    <w:ilvl w:val="0"/>
                    <w:numId w:val="9"/>
                </w:numPr>
            </wpPr>
            <w:r>
                <w:t>this text in numbered or bulleted list
                </w:t>
            </w:r>
        </w:p>

        bullet preceded by one tab for every indentation level.
        """
        try:
            pPr = next(paragraph.iterfind(qn("w:pPr")))
            numPr = next(pPr.iterfind(qn("w:numPr")))
            numId = next(numPr.iterfind(qn("w:numId"))).attrib[qn("w:val")]
            ilvl = next(numPr.iterfind(qn("w:ilvl"))).attrib[qn("w:val")]
            try:
                numFmt = self.numId2numFmts[str(numId)][int(ilvl)]
            except IndexError:
                # give up and put a bullet
                numFmt = "bullet"
        except (StopIteration, KeyError):
            # not a numbered paragraph
            return ""

        def format_bullet(bullet: str) -> str:
            """Indent, format and pad the bullet or number string.

            :param bullet: any kind of list-item string (bullet, number, Roman, ...)
            :return: formatted bullet string
            """
            if bullet != nums.bullet():
                bullet += ")"
            return "\t" * int(ilvl) + bullet + "\t"

        number = _increment_list_counter(self.numId2count[numId], str(ilvl))
        get_unformatted_bullet_str = _get_bullet_function(numFmt)
        return format_bullet(get_unformatted_bullet_str(number))



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['docx2python']

package_data = \
{'': ['*']}

install_requires = \
['lxml>=4.9.2,<5.0.0']

setup_kwargs = {
    'name': 'docx2python',
    'version': '2.6.0',
    'description': 'Extract content from docx files',
    'long_description': '# docx2python\n\nExtract docx headers, footers, text, footnotes, endnotes, properties, and images to a Python object.\n\n`README_DOCX_FILE_STRUCTURE.md` may help if you\'d like to extend docx2python.\n\nFor a summary of what\'s new in docx2python 2, scroll down to **New in docx2python Version 2**\n\nThe code is an expansion/contraction of [python-docx2txt](https://github.com/ankushshah89/python-docx2txt) (Copyright (c) 2015 Ankush Shah). The original code is mostly gone, but some of the bones may still be here.\n\n__shared features__:\n* extracts text from docx files\n* extracts images from docx files\n\n__additions:__\n* extracts footnotes and endnotes\n* converts bullets and numbered lists to ascii with indentation\n* converts hyperlinks to ``<a href="http:/...">link text</a>``\n* retains some structure of the original file (more below)\n* extracts document properties (creator, lastModifiedBy, etc.)\n* inserts image placeholders in text (``\'----image1.jpg----\'``)\n* inserts plain text footnote and endnote references in text (``\'----footnote1----\'``)\n* (optionally) retains font size, font color, bold, italics, and underscore as html\n* extract math equations\n* extract user selections from checkboxes and dropdown menus\n\n__subtractions:__\n* no command-line interface\n* will only work with Python 3.8+\n\n\n## Installation\n```bash\npip install docx2python\n```\n\n## Use\n\n``` python\nfrom docx2python import docx2python\n\n# extract docx content\nwith docx2python(\'path/to/file.docx\') as docx_content:\n    print(docx_content.text)\n\ndocx_content = docx2python(\'path/to/file.docx\')\nprint(docx_content.text)\ndocx_content.close()\n\n# extract docx content, write images to image_directory\nwith docx2python(\'path/to/file.docx\', \'path/to/image_directory\') as docx_content:\n    print(docx_content.text)\n\n# extract docx content with basic font styles converted to html\nwith docx2python(\'path/to/file.docx\', html=True) as docx_content:\n    print(docx_content.text)\n```\n\n`docx2python` opens a zipfile object and (lazily) reads it. Use context management (`with ... as`) to close this zipfile object or explicitly close with `docx_content.close()`.\n\nNote on html feature:\n* supports ``<i>``italic, ``<b>``bold, ``<u>``underline, ``<s>``strike, ``<sup>``superscript, ``<sub>``subscript, ``<span style="font-variant: small-caps">``small caps, ``<span style="text-transform:uppercase">``all caps, ``<span style="background-color: yellow">``highlighted, ``<span style="font-size:32">``font size, ``<span style="color:#ff0000">``colored text.\n* hyperlinks will always be exported as html (``<a href="http:/...">link text</a>``), even if ``html=False``, because I couldn\'t think of a more canonical representation.\n* every tag open in a paragraph will be closed in that paragraph (and, where appropriate, reopened in the next paragraph). If two subsequenct paragraphs are bold, they will be returned as `<b>paragraph a</b>`, `<b>paragraph b</b>`. This is intentional to make  each paragraph its own entity.\n* if you specify `html=True`, `&`, `>` and `<` in your docx text will be encoded as `&amp`, `&gt;` and `&lt;`\n\n## Return Value\n\nFunction `docx2python` returns a DocxContent instance with several attributes.\n\n__header__ - contents of the docx headers in the return format described herein\n\n__footer__ - contents of the docx footers in the return format described herein\n\n__body__ - contents of the docx in the return format described herein\n\n__footnotes__ - contents of the docx in the return format described herein\n\n__endnotes__ - contents of the docx in the return format described herein\n\n__document__ - header  + body + footer (read only)\n\n__text__ - all docx text as one string, similar to what you\'d get from `python-docx2txt`\n\n__properties__ - docx property names mapped to values (e.g., `{"lastModifiedBy": "Shay Hill"}`)\n\n__images__ - image names mapped to images in binary format. Write to filesystem with\n\n```\nfor name, image in result.images.items():\n    with open(name, \'wb\') as image_destination:\n        write(image_destination, image)\n\n# or\n\nwith docx2python(\'path/to/file.docx\', \'path/to/image/directory\') as docx_content:\n    ...\n\n# or\n\nwith docx2python(\'path/to/file.docx\') as docx_content:\n    docx_content.save_images(\'path/to/image/directory\')\n\n```\n\n__docx_reader__ - a DocxReader (see `docx_reader.py`) instance with several methods for extracting xml portions.\n\n\n## Arguments\n\n    def docx2python(\n        docx_filename: str | Path | BytesIO,\n        image_folder: str | None = None,\n        html: bool = False,\n        paragraph_styles: bool = False,\n        extract_image: bool | None = None,\n        duplicate_merged_cells: bool = False\n    ) -> DocxContent:\n        """\n        Unzip a docx file and extract contents.\n\n        :param docx_filename: path to a docx file\n        :param image_folder: optionally specify an image folder\n            (images in docx will be copied to this folder)\n        :param html: bool, extract some formatting as html\n        :param paragraph_styles: prepend the paragraphs style (if any, else "") to each\n            paragraph. This will only be useful with ``*_runs`` attributes.\n        :param duplicate_merged_cells: bool, duplicate merged cells to return a mxn\n            nested list for each table (default False)\n        :return: DocxContent object\n        """\n\n\n## Return Format\n\nSome structure will be maintained. Text will be returned in a nested list, with paragraphs always at depth 4 (i.e., `output.body[i][j][k][l]` will be a paragraph).\n\nIf your docx has no tables, output.body will appear as one a table with all content in one cell:\n\n```python\n[  # document\n    [  # table\n        [  # row\n            [  # cell\n                "Paragraph 1",\n                "Paragraph 2",\n                "-- bulleted list",\n                "-- continuing bulleted list",\n                "1)  numbered list",\n                "2)  continuing numbered list"\n                "    a)  sublist",\n                "        i)  sublist of sublist",\n                "3)  keeps track of indention levels",\n                "    a)  resets sublist counters"\n            ]\n        ]\n     ]\n ]\n```\n\nTable cells will appear as table cells. Text outside tables will appear as table cells.\n\n\nA docx document can be tables within tables within tables. Docx2Python flattens most of this to more easily navigate\nwithin the content.\n\n## Working with output\n\nThis package provides several documented helper functions in [the ``docx2python.iterators`` module](https://docx2python.readthedocs.io/en/latest/docx2python.html#module-iterators). Here are a few recipes possible with these functions:\n\n```python\nfrom docx2python.iterators import enum_cells\n\ndef remove_empty_paragraphs(tables):\n    for (i, j, k), cell in enum_cells(tables):\n        tables[i][j][k] = [x for x in cell if x]\n```\n\n```\n>>> tables = [[[[\'a\', \'b\'], [\'a\', \'\', \'d\', \'\']]]]\n>>> remove_empty_paragraphs(tables)\n    [[[[\'a\', \'b\'], [\'a\', \'d\']]]]\n```\n\n```python\nfrom docx2python.iterators import enum_at_depth\n\ndef html_map(tables) -> str:\n    """Create an HTML map of document contents.\n\n    Render this in a browser to visually search for data.\n\n    :tables: value could come from, e.g.,\n        * docx_to_text_output.document\n        * docx_to_text_output.body\n    """\n\n    # prepend index tuple to each paragraph\n    for (i, j, k, l), paragraph in enum_at_depth(tables, 4):\n        tables[i][j][k][l] = " ".join([str((i, j, k, l)), paragraph])\n\n    # wrap each paragraph in <pre> tags\n    for (i, j, k), cell in enum_at_depth(tables, 3):\n        tables[i][j][k] = "".join(["<pre>{x}</pre>".format(x) for x in cell])\n\n    # wrap each cell in <td> tags\n    for (i, j), row in enum_at_depth(tables, 2):\n        tables[i][j] = "".join(["<td>{x}</td>".format(x) for x in row])\n\n    # wrap each row in <tr> tags\n    for (i,), table in enum_at_depth(tables, 1):\n        tables[i] = "".join("<tr>{x}</tr>".format(x) for x in table)\n\n    # wrap each table in <table> tags\n    tables = "".join([\'<table border="1">{x}</table>\'.format(x) for x in tables])\n\n    return ["<html><body>"] + tables + ["</body></html>"]\n```\n\n```\n>>> tables = [[[[\'a\', \'b\'], [\'a\', \'d\']]]]\n>>> html_map(tables)\n<html>\n    <body>\n        <table border="1">\n            <tr>\n                <td>\n                    \'(0, 0, 0, 0) a\'\n                    \'(0, 0, 0, 1) b\'\n                </td>\n                <td>\n                    \'(0, 0, 1, 0) a\'\n                    \'(0, 0, 1, 1) d\'\n                </td>\n            </tr>\n        </table>\n    </body>\n</html>\n```\n\n[See helper functions.](https://docx2python.readthedocs.io/en/latest/index.html)\n\nSome fine print about checkboxes:\n\nMS Word has checkboxes that can be checked any time, and others that can only be checked when the form is locked.\nThe previous print as. ``\\u2610`` (open checkbox) or ``\\u2612`` (crossed checkbox). Which this module, the latter will\ntoo. I gave checkboxes a bailout value of ``----checkbox failed----`` if the xml doesn\'t look like I expect it to,\nbecause I don\'t have several-thousand test files with checkboxes (as I did with most of the other form elements).\nCheckboxes *should* work, but please let me know if you encounter any that do not.\n\n# New in docx2python Version 2\n\n## merge consecutive runs with identical formatting\n\nMS Word will break up text runs arbitrarily, often in the middle of a word.\n\n\n    <w:r>\n        <w:t>work to im</w:t>\n    </w:r>\n    <w:r>\n        <w:t>prove docx2python</w:t>\n    </w:r>\n\nThis makes things like algorithmic search-and-replace problematic. Docx2python does not currently write docx files,\nbut I often use docx templates with placeholders (e.g., `#CATEGORY_NAME#`) then replace those placeholders with data.\nThis won\'t work if your placeholders are broken up (e.g, `#CAT`, `E`, `GORY_NAME#`).\n\nDocx2python v1 merges such runs together when exporting text. Docx2python v2 will merge such runs in the XML as a\npre-processing step. This will allow saving such "repaired" XML later on.\n\n## merge consecutive links with identical hrefs\n\nMS Word will break up links, giving each link a different `rId`, even when these `rIds` point to the same address.\n\n    <w:hyperlink r:id="rId13">  # rID13 points to https://github.com/ShayHill/docx2python\n        <w:r>\n            <w:t>docx2py</w:t>\n        </w:r>\n    </w:hyperlink>\n    <w:hyperlink r:id="rId14">  # rID14 ALSO points to https://github.com/ShayHill/docx2python\n        <w:r>\n            <w:t>thon</w:t>\n        </w:r>\n    </w:hyperlink>\n\nThis is similar to the broken-up runs, but the cause is a little deeper in. Docx2python v1 makes a mess of these.\n\n    <a href="https://github.com/ShayHill/docx2python">docx2py</a>\n    <a href="https://github.com/ShayHill/docx2python">thon</a>\n\nDocx2python v2 will merge such links together in the XML as a pre-processing step. As above, this will allow saving\nsuch "repaired" XML later on.\n\n## correctly handle nested paragraphs\n\nMS Word will nest paragraphs\n\n    <w:p>\n        <w:r>\n            <w:t>text</w:t>\n        </w:r>\n        <w:p>  # paragraph inside a paragraph\n            <w:r>\n                <w:t>text</w:t>\n            </w:r>\n        </w:p>\n        <w:r>\n            <w:t>text</w:t>\n        </w:r>\n    </w:p>\n\nI haven\'t been able to create such a paragraph, but I\'ve found a few files that have them. Docx2pyhon v1 will omit\nclosing html tags when a new paragraph is opened before the old paragraph is closed.\n\n    <b>outer par bold text\n\n    <i>This text is in nested par (not bold)</i>\n\n    outer par bold text</b>\n\nDocx2python v2 will correctly handle such cases, but this will require substantial internal changes to the way\ndocx2python opens and closes paragraphs.\n\n    <b>outer par bold text</b>\n\n    <i>This text is in nested par (not bold)</i>\n\n    </b>outer par bold text</b>\n\n## paragraph styles\n\nThe internal changes allow for easy access to paragraph styles (e.g., `Heading 1`). Docx2python v1 ignores these, even\nwith `html=True`. Docx2python v2 will capture paragraph styles.\n\n    <h1>h1 is a paragraph style<b>bold is a run style</b></h1>\n\n## export xml\n\nTo allow above-described light editing (e.g., search and replace), docx2python v2 will give the user access to\n\n    1. extracted xml files\n    2. the functions used to write these files to a docx\n\nThe user can only go so far with this. A docx file is built from folders full of xml files. None of these xml\nfiles are self contained. But search and replace is enough to make document templates (documents with placeholders for\ndata), and that\'s pretty useful in itself.\n\n## expose some intermediate functionality\n\nNavigating through XML is straightforward with `lxml`. It is a separate step to take whatever you find and bring it\n*out* of the XML. For instance, you may want to iterate over a document, looking for paragraphs with a particular\nformat, then pull the text out of those paragraphs. Docx2python v1 did not separate or expose "iter the document" and\n"pull the content". Docx2python v2 separates and exposes these steps. This will allow easier extension.\n\nSee the `docx_reader.py` module and simple examples in the `utilities.py` module.\n\n## see utilities.py for examples of major new features.\n',
    'author': 'Shay Hill',
    'author_email': 'shay_public@hotmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
